#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Pyaerocom GriddedData class
"""

import os
from collections import OrderedDict as od
import iris
from iris.analysis.cartography import area_weights
from iris.analysis import MEAN
from iris.exceptions import UnitConversionError
from pandas import Timestamp, Series
import numpy as np
import pandas as pd
from pyaerocom import const, logger, print_log

from pyaerocom.exceptions import (CoordinateError,
                                  DataDimensionError,
                                  DataExtractionError,
                                  DimensionOrderError,
                                  TemporalResolutionError,
                                  VariableDefinitionError,
                                  VariableNotFoundError)

from pyaerocom.time_config import IRIS_AGGREGATORS, TS_TYPE_TO_NUMPY_FREQ

from pyaerocom.helpers import (get_time_rng_constraint,
                               get_lon_rng_constraint,
                               get_lat_rng_constraint,
                               cftime_to_datetime64,
                               str_to_iris,
                               to_pandas_timestamp,
                               datetime2str,
                               isrange, isnumeric,
                               delete_all_coords_cube,
                               copy_coords_cube)
from pyaerocom.mathutils import closest_index
from pyaerocom.stationdata import StationData
from pyaerocom.region import Region
from pyaerocom.vert_coords import AltitudeAccess


class GriddedData(object):
    """Base class representing model data
    
    This class is largely based on the :class:`iris.Cube` object. However, this
    object comes with an expanded functionality for convenience, for instance, 
    netCDF files can directly be loaded in the :class:`GriddedData` object,
    whereas :class:`iris.cube.Cube` instances are typically created using
    helper methods such as
    
    1. :func:`iris.load` (returns 
    :class:`iris.cube.CubeList`, i.e. a list-like iterable object that contains 
    instances of :class:`Cube` objects, one for each variable) or 
    
    2. :func:`iris.load_cube` which directly returns a :class:`iris.cube.Cube` 
    instance and typically requires specification of a variable constraint.
    
    The :class:`GriddedData` object represents one variable in space and time, as
    well as corresponding meta information. Since it is based on the https://github.com/SciTools/iris/issues/1977
    :class:`iris.cube.Cube` it is optimised for netCDF files that follow the
    CF conventions and may not work for files that do not follow this standard.
       
    Parameters
    ----------
    input : :obj:`str:` or :obj:`Cube`
        data input. Can be a single .nc file or a preloaded iris Cube.
    var_name : :obj:`str`, optional
        variable name that is extracted if `input` is a file path. Irrelevant
        if `input` is preloaded Cube

    Example
    -------
    >>> from pyaerocom.io.testfiles import get
    >>> files = get()
    >>> data = GriddedData(files['models']['aatsr_su_v4.3'], 
    ...  ot self.grid.var_name is None:
                const.print_log.warning('Overwriting existing variable name {} '
                                        'in with {}'.format(self.grid.var_name, 
                                                 var_name))                  var_name="od550aer")
    >>> print(data.var_name)
    od550aer
    >>> print(type(data.longitude))
    <class 'iris.coords.DimCoord'>
    >>> print(data.longitude.points.min(), data.longitude.points.max())
    -179.5 179.5
    >>> print(data.latitude.points.min(), data.latitude.points.max())
    -89.5 89.5
    >>> print(data.time.points.min(), data.time.points.max())
    0.0 365.0
    >>> tstamps = data.time_stamps()
    >>> print(tstamps[0], tstamps[-1])
    2008-01-01T00:00:00.000000 2008-12-31T00:00:00.000000
    >>> data_cropped = data.crop(lat_range=(-60, 60), lon_range=(160, 180),
    ...                          time_range=("2008-02-01", "2008-02-15"))
    >>> print(data_cropped.shape)
    (15, 120, 20)
    
    Attributes
    ----------
    grid
        underlying data type (hopefully :class:`iris.cube.Cube` in most cases)
    """
    _grid = None
    _GRID_IO = const.GRID_IO
    #: Req. order of dimension coordinates for time-series computation
    COORDS_ORDER_TSERIES = ['time', 'latitude', 'longitude']
    _MAX_SIZE_GB = 64 #maximum file size for in-memory operations

    SUPPORTED_VERT_SCHEMES = ['mean', 'max', 'min', 'surface', 'altitude',
                              'profile']
    
    _META_ADD = od(from_files         = [],
                   data_id            = "n/d",
                   var_name_read      = "n/d",
                   ts_type            = "n/d",
                   regridded          = False,
                   outliers_removed   = False,
                   computed           = False,
                   concatenated       = False,
                   region             = None,
                   reader             = None)
    
    def __init__(self, input=None, var_name=None, convert_unit_on_init=True,
                 **meta):
        
        if input is None:
            input = iris.cube.Cube([])
            
        self._grid = None
        self._reader = None
        #attribute used to store area weights (if applicable, see method
        #area_weights)
        self._area_weights = None
        self._altitude_access = None
        
        # list of coordinate names as returned by name() method of iris coordinate
        # will be filled upon access of coord_names
        self._coord_names = None 
        # list of containing var_name attributes of all coordinates
        self._coord_var_names = None
        self._coord_standard_names = None
        self._coord_long_names = None
        
        if input:
            self.load_input(input, var_name)
        
        self.update_meta(**meta)
        
        if self.has_data and self.var_name is not None and convert_unit_on_init:       
            self.check_unit()
    
    @property
    def var_name(self):
        """Name of variable"""
        return self.grid.var_name
    
    @var_name.setter
    def var_name(self, val):
        """Name of variable"""
        self.grid.var_name = val
    
    @property
    def var_name_aerocom(self):
        """AeroCom variable name"""
        try:
            return const.VARS[self.var_name].var_name
        except:
            return None
    
    @property
    def var_info(self):
        """Print information about variable"""
        if not self.var_name in const.VARS:
            try:
                return const.VARS[self.var_name_aerocom]
            except:
                raise VariableDefinitionError('No default access available for '
                                              'variable {}'.format(self.var_name))
        return const.VARS[self.var_name]
    
    @property
    def ts_type(self):
        """Temporal resolution"""
        if self.metadata['ts_type'] == 'n/d':
            self.infer_ts_type()
        
        return self.metadata['ts_type']
    
    @property
    def standard_name(self):
        """Standard name of variable"""
        return self.grid.standard_name
    
    @property
    def long_name(self):
        """Long name of variable"""
        return self.grid.long_name
    
    @long_name.setter
    def long_name(self, val):
        self.grid.long_name = val
            
    @property
    def unit_ok(self):
        """Boolean specifying if variable unit is AeroCom default"""
        try:
            var = const.VARS[self.cube.var_name]
            if var.has_unit and var.units == self.units:
                return True
            return False
        except:
            return False
      
    @property
    def suppl_info(self):
        w = DeprecationWarning('Outdated attribute suppl_info. Please use '
                               'metadata instead')
        const.print_log.warning(w)
        return self.metadata
    
    @property
    def metadata(self):
        return self.cube.attributes
    
    def check_unit(self):
        """Check if unit of data is AeroCom default and convert if not
        """
        try:
            if not self.unit_ok:
                var = const.VARS[self.cube.var_name]
                logger.info('Attempting unit conversion from {} to {}'
                            .format(self.units, var.units))
                self.convert_unit(var.units)
        except (VariableDefinitionError, UnitConversionError,
                MemoryError, ValueError) as e:
            logger.info('Failed to convert unit. Reason: {}'.format(repr(e)))
            
    @property
    def data_revision(self):
        """Revision string from file Revision.txt in the main data directory
        """
        if self.from_files:
            data_dir = os.path.dirname(self.from_files[0])
            revision_file = os.path.join(data_dir, const.REVISION_FILE)
            if os.path.isfile(revision_file):
                with open(revision_file, 'rt') as in_file:
                    revision = in_file.readline().strip()
                    in_file.close()
    
                return revision
        return 'n/a'
        
    @property
    def reader(self):
        """Instance of reader class from which this object was created"""
        r = self._reader
        from pyaerocom.io import ReadGridded
        if not isinstance(r, ReadGridded):
            self._reader = r = ReadGridded(self.data_id)
        return r

    @reader.setter
    def reader(self, val):
        self._reader = val
        
    @property
    def concatenated(self):
        return self.metadata['concatenated']

    @property
    def computed(self):
        return self.metadata['computed']
    
    

    @property
    def units(self):
        """Unit of data"""
        return self.grid.units
    
    @units.setter
    def units(self, val):
        self.grid.units = val
    
    @property
    def data(self):
        """Data array (n-dimensional numpy array)
        
        Note
        ----
        This is a pointer to the data object of the underlying iris.Cube
        instance and will load the data into memory. Thus, in case of large
        datasets, this may lead to a memory error
        """
        return self.grid.data
    
    @data.setter
    def data(self, array):
        if not isinstance(array, np.ndarray):
            raise ValueError('Cannot set data array: need numpy.ndarray')
        elif not array.shape == self.grid.data.shape:
            raise DataDimensionError('Cannot assign dataarray: shape mismatch. '
                                     'Got: {}, Need: {}'
                                     .format(array.shape,self.grid.shape))
        self.grid.data = array
    
    @property
    def altitude_access(self):
        if not isinstance(self._altitude_access, AltitudeAccess):
            self._altitude_access = AltitudeAccess(self)    
        return self._altitude_access
    
    @property
    def delta_t(self):
        """Array containing timedelta values for each time stamp"""
        ts = self.time_stamps()
        if len(ts) < 2:
            raise AttributeError('Need at least 2 timestamps in GriddedData in '
                                 'order to compute delta-t')
        return (ts[1:] - ts[0:-1])
    
    def check_frequency(self):
        """Check if all datapoints are sampled at the same time frequency"""
        dt = np.unique(self.delta_t)
        if len(dt) > 1:
            raise AttributeError('Irregular time-frequency')
        freq = TS_TYPE_TO_NUMPY_FREQ[self.ts_type]
        if not int(dt.astype('timedelta64[{}]'.format(freq))) == 1:
            raise AttributeError('Mismatch between sampling freq and '
                                 'actual frequency of values in time dimension ')
            
    def infer_ts_type(self):
        """Try to infer sampling frequency from time dimension data
        
        Returns
        -------
        str
            ts_type that was inferred (is assigned to metadata too)
            
        Raises 
        ------
        DataDimensionError
            if data object does not contain a time dimension
        """
        if not self.has_time_dim:
            raise DataDimensionError('Cannot infer frequency. Data has no time '
                                     'dimension')
        dt = np.unique(self.delta_t)
        if len(dt) > 1:
            raise ValueError('Could not identify unique frequency')
        dt = dt[0]
        for ts_type, freq in TS_TYPE_TO_NUMPY_FREQ.items():
            val = dt.astype('timedelta64[{}]'.format(freq)).astype(int)
            if val == 1:
                self.metadata['ts_type'] = ts_type
                return ts_type
        raise AttributeError('Failed to infer ts_type from data')
        
        
    @property
    def TS_TYPES(self):
        """List with valid filename encryptions specifying temporal resolution
        """
        return self.io_opts.GRID_IO.TS_TYPES
      
    @property
    def from_files(self):
        """List of file paths from which this data object was created"""
        return self.metadata['from_files']
    
    @property
    def is_masked(self):
        """Flag specifying whether data is masked or not
        
        Note
        ----
        This method only works if the data is loaded.
        """
        if self.grid.has_lazy_data():
            raise AttributeError("Information cannot be accessed. Data is not "
                                 "available in memory (lazy loading)")
        return isinstance(self.grid.data, np.ma.core.MaskedArray)
    
    @property
    def base_year(self):
        """Base year of time dimension
        
        Note
        ----
        Changing this attribute will update the time-dimension.
        """
        if not self.has_time_dim:
            raise DataDimensionError('Could not access base year: data has no '
                                     'time dimension')
        try:
            return self.time.units.utime().origin.year
        except Exception as e:
            raise DataDimensionError('Could access base-year. Unexpected error: '
                                     '{}'.format(repr(e)))
            
    @base_year.setter
    def base_year(self, val):
        self.change_base_year(val)
            
    def change_base_year(self, new_year):
        """Changes base year of time dimension 
        
        Relevant, e.g. for climatological analyses
        
        Parameters
        -----------
        new_year : int 
            new base year (can also be other than integer if it is convertible)
        """
        if not self.has_time_dim:
            raise DataDimensionError('Data object has no time dimension ... ')
        if isinstance(new_year, str):
            try:
                new_year = int(new_year)
                if not new_year > -2000 and new_year < 20000:
                    raise ValueError('Need value between -2000 and 20000')
            except Exception as e:
                raise ValueError(repr(e))
        from cf_units import Unit
        u = self.time.units
        
        self.time.units = Unit(u.origin.replace(str(u.utime().origin.year), 
                                                str(new_year)), 
                                calendar=u.calendar)
    @property
    def start(self):
        """Start time of dataset as datetime64 object"""
        if not self.has_time_dim:
            raise ValueError('GriddedData has no time dimension')
        t = cftime_to_datetime64(self.time[0])[0]
        
        try:
            dtype_appr = 'datetime64[{}]'.format(TS_TYPE_TO_NUMPY_FREQ[self.ts_type])
            t=t.astype(dtype_appr)
        except:
            logger.exception('Failed to round start time {} to beginning of '
                             'frequency {}'.format(t, self.ts_type))
        return t.astype('datetime64[us]')

    
    @property
    def stop(self):
        """Start time of dataset as datetime64 object"""
        if not self.has_time_dim:
            raise ValueError('GriddedData has no time dimension')
        t = cftime_to_datetime64(self.time[-1])[0]
        try:
            freq = TS_TYPE_TO_NUMPY_FREQ[self.ts_type]
            dtype_appr = 'datetime64[{}]'.format(freq)
            
            t = t.astype(dtype_appr) + np.timedelta64(1, unit=freq)
            t = t.astype('datetime64[us]') - np.timedelta64(1,unit='us')
            return t
        except:
            logger.exception('Failed to round start time {} to beggining of '
                             'frequency {}'.format(t, self.ts_type))
            return t.astype('datetime64[us]')
    
    @property
    def cube(self):
        """Instance of underlying cube object"""
        return self.grid
    
    @cube.setter
    def cube(self, val):
        """Instance of underlying cube object"""
        self.grid = val
    
    @property
    def grid(self):
        """Underlying grid data object"""
        return self._grid
    
    @grid.setter
    def grid(self, value):
        if not isinstance(value, iris.cube.Cube):
            raise TypeError("Grid data format %s is not supported, need Cube" 
                            %type(value))

        for key, val in self._META_ADD.items():
            if not key in value.attributes:
                value.attributes[key] = val
        self._grid = value
        
    @property
    def plot_settings(self):
        """:class:`Variable` instance that contains plot settings
        
        The settings can be specified in the variables.ini file based on the
        unique var_name, see e.g. `here <http://aerocom.met.no/pyaerocom/
        config_files.html#variables>`__
        
        If no default settings can be found for this variable, all parameters
        will be initiated with ``None``, in which case the Aerocom plot method
        uses
        """
        return const.VARS[self.var_name]
            
    @property 
    def name(self):
        """ID of model to which data belongs"""
        logger.warning('Deprecated attribute name, please use data_id instead')
        return self.metadata["data_id"]
    
    @property
    def data_id(self):
        """ID of data object (e.g. model run ID, obsnetwork ID)
        
        Note
        ----
        This attribute was formerly named ``name`` which is alse the 
        corresponding attribute name in :attr:`metadata`
        """
        return self.metadata['data_id']
    
    @property
    def is_climatology(self):
        ff = self.from_files
        if len(ff) == 1 and '9999' in os.path.basename(ff[0]):
            return True
        return False

    @property
    def has_data(self):
        """True if sum of shape of underlying Cube instance is > 0, else False
        """
        return True if bool(sum(self._grid.shape)) else False
        
    @property
    def shape(self):
        return self._grid.shape 
    
    @property
    def lon_res(self):
        if not 'longitude' in self:
            raise AttributeError('Data does not contain longitude information')
        vals = np.diff(self.longitude.points)
        val = vals.mean()
        if vals.std() / val > 0.0001:
            raise ValueError('Check longitudes')
        return val
    
    @property
    def lat_res(self):
        if not 'latitude' in self:
            raise AttributeError('Data does not contain longitude information')
        vals = np.diff(self.latitude.points)
        val = vals.mean()
        if vals.std() / val > 0.0001:
            raise ValueError('Check latitudes')
        return val
    
    @property
    def ndim(self):
        """Number of dimensions"""
        if not self.has_data:
            raise NotImplementedError("No data available...")
        return self.grid.ndim
    
    @property
    def coords_order(self):
        """Array containing the order of coordinates"""
        return self.coord_names    
    
    @property
    def coord_names(self):
        """List containing coordinate names"""
        if not self.has_data:
            return []
        elif self._coord_names is None:
            self._update_coord_info()
        return self._coord_names
    
    @property
    def dimcoord_names(self):
        """List containing coordinate names"""
        if not self.has_data:
            return []
        return [c.name() for c in self.grid.dim_coords]
        
    @property
    def area_weights(self):
        """Area weights of lat / lon grid"""
        if self._area_weights is None:
            self.calc_area_weights()
        return self._area_weights
    
    @area_weights.setter
    def area_weights(self, val):
        raise AttributeError("Area weights cannot be set manually yet...")
     
    @property
    def has_latlon_dims(self):
        """Boolean specifying whether data has latitude and longitude dimensions"""
        return all([dim in self.dimcoord_names for dim in ['latitude',
                                                           'longitude']])
    @property
    def has_time_dim(self):
        """Boolean specifying whether data has latitude and longitude dimensions"""
        return 'time' in self.dimcoord_names
    
    def load_input(self, input, var_name=None, perform_fmt_checks=None):
        """Import input as cube
        
        Parameters
        ----------
        input : :obj:`str:` or :obj:`Cube`
            data input. Can be a single .nc file or a preloaded iris Cube.
        var_name : :obj:`str`, optional
            variable name that is extracted if `input` is a file path . Irrelevant
            if `input` is preloaded Cube
        perform_fmt_checks : bool, optional
            perform formatting checks based on information in filenames
    
        """
        if isinstance(input, iris.cube.Cube):
            self.grid = input #instance of Cube
            
        elif isinstance(input, str) and os.path.exists(input):
            from pyaerocom.io.iris_io import load_cube_custom
            self.grid = load_cube_custom(input, var_name,
                                         perform_fmt_checks=perform_fmt_checks)
            self.metadata["from_files"].append(input)
            try:
                from pyaerocom.io.helpers import get_metadata_from_filename
                self.update_meta(**get_metadata_from_filename(input))
            except:
                logger.warning('Failed to access metadata from filename')
        else:
            raise IOError('Failed to load input: {}'.format(input))
        
        if var_name is not None and self.var_name != var_name: 
            self.var_name = var_name
            
     
    def _get_info_from_filenames(self):
        """Try access AeroCom meta info from filenames assigned to this object
        """
        raise NotImplementedError
        from pyaerocom.io import FileConventionRead
        c = FileConventionRead(from_file=self.from_files[0])
        info = c.get_info_from_file(self.from_files[0])
        for f in self.from_files[1:]:
            add_info = f.from_file(f)
            
        
    def convert_unit(self, new_unit):
        """Convert unit of data to new unit"""
        if self._size_GB > self._MAX_SIZE_GB:
            raise MemoryError('Cannot convert unit in {} since data is too '
                              'large ({} GB)'.format(self.name, self._size_GB))
        self.grid.convert_units(new_unit)
        
    def time_stamps(self):
        """Convert time stamps into list of numpy datetime64 objects
        
        The conversion is done using method :func:`cfunit_to_datetime64`
        
        Returns
        -------
        list 
            list containing all time stamps as datetime64 objects 
        """
        if self.has_time_dim:    
            return cftime_to_datetime64(self.time)

    def check_coord_order(self):
        """Wrapper for :func:`check_dimcoords_tseries`"""
        logger.warning(DeprecationWarning('Method was renamed, please use '
                                          'check_dimcoords_tseries'))
        return self.check_dimcoords_tseries()
    
    def check_dimcoords_tseries(self):
        """Check order of dimension coordinates for time series retrieval
        
        For computation of time series at certain lon / lat coordinates, the 
        data dimensions have to be in a certain order specified by 
        :attr:`COORDS_ORDER_TSERIES`.
        
        This method checks the current order (and dimensionality) of data and
        raises appropriate errors.
        
        Raises
        ------
        DataDimensionError
            if dimension of data is not supported (currently, 3D or 4D data
            is supported)
        DimensionOrderError
            if dimensions are not in the right order (in which case 
            :func:`reorder_dimensions_tseries` may be used to catch the 
            Exception)
        """
        order = self.COORDS_ORDER_TSERIES
        if not self.ndim in (3,4):
            raise DataDimensionError('Time series extraction requires at least 3 '
                            'coordinates in cube')
        check = self.dimcoord_names
        if not len(check) >= 3:
            raise DataDimensionError('One of the data dimension coordinates '
                                     'may not be defined')
            
        for i, item in enumerate(check[:3]):
            if not item == order[i]:
                raise DimensionOrderError('Invalid order of grid '
                                          'dimension, need {}, got {}'
                                          .format(order,
                                                  check))
        
            
    def reorder_dimensions_tseries(self):
        """Reorders dimensions of data such that :func:`to_time_series` works
        """
        order = self.COORDS_ORDER_TSERIES
        new_order = []
        coord_names = [c.name() for c in self.grid.dim_coords]
        for coord_name in order:
            new_order.append(coord_names.index(coord_name))
        if not len(new_order) == self.ndim:
            for i in range(self.ndim):
                if not i in new_order:
                    new_order.append(i)
        self.transpose(new_order)
        self.check_dimcoords_tseries()
        
    def transpose(self, new_order):
        """Re-order data dimensions in object
        
        Wrapper for :func:`iris.cube.Cube.transpoe`
        
        Note
        ----
        Changes THIS object (i.e. no new instance of :class:`GriddedData` will 
        be created)
        
        Parameters
        ----------
        order : list
            new index order
        """
        self.grid.transpose(new_order)
        
    def to_time_series(self, sample_points=None, scheme="nearest", 
                       collapse_scalar=True, vert_scheme=None, 
                       add_meta=None, **coords):

        """Extract time-series for provided input coordinates (lon, lat)

        Extract time series for each lon / lat coordinate in this cube or at
        predefined sample points (e.g. station data). If sample points are
        provided, the cube is interpolated first onto the sample points.

        Todo
        ----
        Check Memory error handle
        
        Parameters
        ----------
        sample_points : list
            coordinates (e.g. lon / lat) at which time series is supposed to be
            retrieved
        scheme : str or iris interpolator object
            interpolation scheme (for details, see :func:`interpolate`)
        collapse_scalar : bool
            see :func:`interpolate`
        vert_scheme : str
            string specifying how to treat vertical coordinates. This is only
            relevant for data that contains vertical levels. It will be ignored
            otherwise. Note that if the input coordinate specifications contain
            altitude information, this parameter will be set automatically to
            'altitude'. Allowed inputs are all data collapse schemes that 
            are supported by :func:`pyaerocom.helpers.str_to_iris` (e.g. `mean, 
            median, sum`). Further valid schemes are `altitude, surface, 
            profile`.
            If not other specified and if `altitude` coordinates are provided
            via sample_points (or **coords parameters) then, vert_scheme will 
            be set to `altitude`. Else, `profile` is used.
        add_meta : dict, optional
            dictionary specifying additional metadata for individual input 
            coordinates. Keys are meta attribute names (e.g. station_name)
            and corresponding values are lists (with length of input coords)
            or single entries that are supposed to be assigned to each station.
            E.g. `add_meta=dict(station_name=[<list_of_station_names>])`). 
        **coords
            additional keyword args that may be used to provide the interpolation
            coordinates (for details, see :func:`interpolate`)

        Returns
        -------
        list
            list of result dictionaries for each coordinate. Dictionary keys
            are: ``longitude, latitude, var_name``
        """
        try:
            self.check_dimcoords_tseries()
        except DimensionOrderError:
            self.reorder_dimensions_tseries()
        # if the method makes it to this point, it is 3 or 4 dimensional 
        # and the first 3 dimensions are time, latitude, longitude.
        
        # init input for sample points
        if sample_points is None:
            sample_points = []
            for c, v in coords.items():
                if isnumeric(v):
                    v= [v]
                sample_points.append((c, v))

        lens = [len(x[1]) for x in sample_points]
        if not all([lens[0]==x for x in lens]):
            raise ValueError("Arrays for sample coordinates must have the "
                             "same lengths")
        if self.ndim == 3: #data does not contain vertical dimension
            return self._to_timeseries_2D(sample_points, scheme,
                                          collapse_scalar,
                                          add_meta=add_meta)
        
        return self._to_timeseries_3D(sample_points, scheme, 
                                      collapse_scalar, vert_scheme,
                                      add_meta=add_meta)
    
        
    def _to_timeseries_2D(self, sample_points, scheme, collapse_scalar,
                          add_meta=None):
        """Extract time-series for provided input coordinates (lon, lat)
        
        Todo
        ----
        Check Memory error handle
        
        Parameters
        ----------
        sample_points : list
            coordinates (e.g. lon / lat) at which time series is supposed to be
            retrieved
        scheme : str or iris interpolator object
            interpolation scheme (for details, see :func:`interpolate`)
        collapse_scalar : bool
            see :func:`interpolate`
            
        Returns
        -------
        list
            list of result dictionaries for each coordinate. Dictionary keys
            are: ``longitude, latitude, var_name``
        """
        if self.ndim != 3:
            raise Exception('Developers: Debug! Users: please contact '
                            'developers :)')
        
        data = self.interpolate(sample_points, scheme, collapse_scalar)
        var = self.var_name
        times = data.time_stamps()
        lats = [x[1] for x in sample_points if x[0] == "latitude"][0]
        lons = [x[1] for x in sample_points if x[0] == "longitude"][0]
        arr = data.grid.data
        grid_lons = data.longitude.points
        result = []
        meta_iter = {}
        meta_glob = {}
        if add_meta is not None:
            for meta_key, meta_val in add_meta.items():
                try:
                    if not len(meta_val) == len(lons):
                        raise ValueError
                    meta_iter[meta_key] = meta_val
                except:
                    meta_glob[meta_key] = meta_val
            
        for i, lat in enumerate(lats):
            lon = lons[i]
            j = np.where(grid_lons==lon)[0][0]
            
            data = StationData(latitude=lat, 
                               longitude=lon,
                               data_id=self.name,
                               ts_type=self.ts_type)
            data.var_info[var] = {'units':self.units}
            vals = arr[:, i, j]
            
            data[var] = Series(vals, index=times)
            for meta_key, meta_val in meta_iter.items():
                data[meta_key] = meta_val[i]
            for meta_key, meta_val in meta_glob.items():
                data[meta_key] = meta_val
            result.append(data)
        return result

    def _apply_vert_scheme(self, sample_points, vert_scheme=None):
        """Helper method that checks and infers vertical scheme for time
        series computation from 3D data (used in :func:`_to_timeseries_3D`)"""        
        
        if vert_scheme is None:
            const.print_log.info('Setting vert_scheme in GriddedData to mean')
            vert_scheme ='mean'        
        try:
            self.check_dimcoords_tseries()
        except DimensionOrderError:
            self.reorder_dimensions_tseries()
        
        cname = self.dimcoord_names[-1]
        
        if not vert_scheme in self.SUPPORTED_VERT_SCHEMES:
            raise ValueError('Invalid input for vert_scheme: {}. Supported '
                             'schemes are: {}'
                             .format(vert_scheme, self.SUPPORTED_VERT_SCHEMES))
        if vert_scheme == 'surface':
            vert_index = self._infer_index_surface_level()
            return self[:,:,:,vert_index]
        elif vert_scheme == 'altitude':
            if not 'altitude' in [sp[0] for sp in sample_points]:
                raise ValueError('Require altitude specification in sample '
                                 'points for vert_scheme altitude')
            if not self.check_altitude_access():
                raise DataDimensionError('Cannot access altitude '
                                         'information')
            raise NotImplementedError('Cannot yet retrieve timeseries at '
                                      'altitude levels. Coming soon...')
        elif vert_scheme == 'profile':
            raise NotImplementedError('Cannot yet retrieve profile timeseries')
        else:
            try:
                # check if vertical scheme can be converted into valid iris 
                # aggregator (in which case vertical dimension is collapsed)
                aggr = str_to_iris(vert_scheme)
            except KeyError:
                pass
            else:
                return self.collapsed(cname, aggr)
            
        raise NotImplementedError('Cannot yet retrieve timeseries '
                                  'from 4D data for vert_scheme {} '
                                  .format(vert_scheme))
    
    def check_altitude_access(self):
        """Checks if altitude levels can be accessed
        
        Returns
        -------
        bool
            True, if altitude access is provided, else False
         
        """
        return self.altitude_access.check_altitude_access()
    
    def get_altitude(self, **coords):
        """Extract (or try to compute) altitude values at input coordinates"""
        if not isinstance(self._altitude_access, AltitudeAccess):
            self.check_altitude_access()
        self._altitude_access.get_altitude(**coords)
        raise NotImplementedError('Coming soon...')
        
    def extract_surface_level(self):
        """Extract surface level from 4D field"""     
        if not self.ndim==4:
            raise DataDimensionError('Can only extract surface level for 4D '
                                     'gridded data object')
        idx = self._infer_index_surface_level()
        return self[:,:,:,idx]
    
    def _infer_index_surface_level(self):
        if not self.ndim == 4:
            raise DataDimensionError('Can only infer surface level for 4D '
                                     'gridded data object')
        try:
            self.check_dimcoords_tseries()
        except DimensionOrderError:
            self.reorder_dimensions_tseries()
        cname = self.dimcoord_names[-1]
        from pyaerocom import vert_coords as vc
        try:
            coord = vc.VerticalCoordinate(cname)
            if coord.lev_increases_with_alt: 
                # vertical coordinate values increase with altitude -> find lowest value (e.g. altitude)
                return np.argmin(self.grid.dim_coords[3].points)
            else:
                # vertical coordinate values decrease with altitude -> find highest value (e.g. pressure)
                return np.argmax(self.grid.dim_coords[3].points)
        except:
            if not const.GRID_IO.INFER_SURFACE_LEVEL:
                raise DataExtractionError('Cannot infer surface level since '
                                          'global option INFER_SURFACE_LEVEL in'
                                          'pyaerocom.const.GRID_IO is deactivated')
            last_lev_idx = self.shape[-1] - 1
            first_lowest_idx = self[0, :, :, 0].data
            first_highest_idx = self[0, :, :, last_lev_idx].data
            if np.nanmean(first_lowest_idx) > np.nanmean(first_highest_idx):
                return 0
            return last_lev_idx
        
    def _to_timeseries_3D(self, sample_points, scheme, collapse_scalar,
                          vert_scheme='surface', add_meta=None):

        # data contains vertical dimension
        data = self._apply_vert_scheme(sample_points, vert_scheme)
    
        # ToDo: check if _to_timeseries_2D can be called here                    
        return data.to_time_series(sample_points, scheme, 
                                   collapse_scalar, add_meta=add_meta)
    
    def to_time_series_single_coord(self, latitude, longitude):
        """Make time series dictionary of single location using neirest coordinate
        
        Todo
        ----
        Crop before extraction
        
        Parameters
        ----------
        latitude : float
            latitude of coordinate
        longitude : float
            longitude of coordinate
            
        Returns
        -------
        dict
            dictionary containing results
        """
        raise NameError(DeprecationWarning('This method is deprecated since '
                                           'version 0.8.0'))
# =============================================================================
#         self.check_dimcoords_tseries()
#         if not self.ndim == 3:
#             raise DataDimensionError('So far, timeseries can only be extracted '
#                                      'from 3 dimensional data...')
#         lons = self.longitude.points
#         lats = self.latitude.points
#         lon_idx = np.argmin(np.abs(lons - longitude))
#         lat_idx = np.argmin(np.abs(lats - latitude))
#         times = self.time_stamps()
#         data = self.grid.data[:, lat_idx, lon_idx]
#         return {'latitude'      : latitude,
#                 'longitude'     : longitude,
#                 'name'          : self.name,
#                 self.var_name   : Series(data, times)}
# =============================================================================
        
    def _closest_time_idx(self, t):
        """Find closest index to input in time dimension"""
        t = self.time.units.date2num(to_pandas_timestamp(t))
        return self.time.nearest_neighbour_index(t)
    
    def find_closest_index(self, **dimcoord_vals):
        """Find the closest indices for dimension coordinate values"""
        idx = {}
        for dim, val in dimcoord_vals.items():
            if not dim in self.coord_names:
                raise DataDimensionError('No such dimension {}'.format(dim))
            elif dim == 'time':
                idx[dim] = self._closest_time_idx(val)
            else:
                idx[dim] = self[dim].nearest_neighbour_index(val)
        return idx
    
    def isel(self, **kwargs):
        raise NotImplementedError('Please use method sel for data selection '
                                  'based on dimension values')
        
    def sel(self, use_neirest=True, **dimcoord_vals):
        """Select subset by dimension names
        
        Note
        ----
        This is a BETA version, please use with care
        
        Parameters
        ----------
        **dimcoord_vals 
            key / value pairs specifying coordinate values to be extracted
            
        Returns
        -------
        GriddedData
            subset data object
        """
        constraints = []
        rng_funs = {'time'   : get_time_rng_constraint,
                    'longitude' : get_lon_rng_constraint,
                    'latitude' : get_lat_rng_constraint}
        
        coord_vals = {}
        for dim, val in dimcoord_vals.items():
            is_rng = isrange(val)
            if is_rng:
                c = rng_funs[dim](val)
                constraints.append(c)
            else:
                if dim == 'time':
                    if isnumeric(val) and val in self['time'].points:
                        _tval = val
                    else:
                        _idx = self._closest_time_idx(val)
                        _tval = self.time[_idx].points[0]
                    _cval = self['time'].units.num2date(_tval)
                    if not use_neirest and _cval != val:
                        raise DataExtractionError('No such value {} in dim {}. '
                                                  'Use option use_neirest to '
                                                  'disregard and extract '
                                                  'neirest neighbour'.format
                                                  (val, dim))
                else:
                    _idx = self[dim].nearest_neighbour_index(val)
                    _cval = self[dim][_idx].points[0]
                    if not use_neirest and _cval != val:
                        raise DataExtractionError('No such value {} in dim {}'
                                                  'Use option use_neirest to '
                                                  'disregard and extract '
                                                  'neirest neighbour'.format
                                                  (val, dim))
                coord_vals[dim] = _cval
                
        if coord_vals:
            constraints.append(iris.Constraint(coord_values=coord_vals))
        
        if len(constraints) > 0:
            c = constraints[0]
            for cadd in constraints[1:]:
                c = c & cadd
        subset = self.extract(c)
        if subset is None:
            raise DataExtractionError('Failed to extract subset for input '
                                      'coordinates {}'.format(dimcoord_vals))
        return subset
                    
    # TODO: Test, confirm and remove beta flag in docstring
    def remove_outliers(self, low=None, high=None):
        """Remove outliers from data

        Parameters
        ----------
        low : float
            lower end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. minimum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        high : float
            upper end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. maximum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        """
        if low is None:
            low = self.var_info.minimum
            print_log.info('Setting {} outlier lower lim: {:.2f}'
                           .format(self.var_name, low))
        if high is None:
            high = self.var_info.maximum
            print_log.info('Setting {} outlier upper lim: {:.2f}'
                           .format(self.var_name, high))
        mask = np.logical_or(self.grid.data < low,
                             self.grid.data > high)
        self.grid.data[mask] = np.nan
        self.metadata['outliers_removed'] = True
        
    def resample_time(self, to_ts_type='monthly'):
        """Downscale in time to predefined resolution resolution
        
        Parameters
        ----------
        to_ts_type : str
            either of the supported temporal resolutions (cf. 
            :attr:`IRIS_AGGREGATORS` in :mod:`helpers`, e.g. "monthly")
        
        Returns
        -------
        GriddedData
            new data object containing downscaled data
            
        Raises
        ------
        TemporalResolutionError
            if input resolution is not provided, or if it is higher temporal 
            resolution than this object
        """
        if not self.has_time_dim:
            raise DataDimensionError('Require time dimension in GriddedData: '
                                     '{}'.format(self.short_str()))
        from pyaerocom.tstype import TsType
        
        to = TsType(to_ts_type)
        current = TsType(self.ts_type)
        
        if current == to:
            logger.info('Data is already in {} resolution'.format(to_ts_type))
            return self
        if not to_ts_type in IRIS_AGGREGATORS:
            raise TemporalResolutionError('Resolution {} cannot '
                'converted'.format(to_ts_type))
        elif current < to: #current resolution is smaller than desired
            raise TemporalResolutionError('Cannot increase '
                'temporal resolution from {} to {}'.format(self.ts_type,
                                          to_ts_type))
        cube = self.grid
        
        # Create aggregators
        aggrs = ['yearly']
        if not to_ts_type in aggrs:
            aggrs.append(to_ts_type)
            
        for aggr in aggrs:
            if not aggr in [c.name() for c in cube.aux_coords]:
                # this adds the corresponding aggregator to the cube
                IRIS_AGGREGATORS[aggr](cube, 'time', name=aggr)
            #IRIS_AGGREGATORS[to_ts_type](cube, 'time', name=to_ts_type)
        # not downscale
        aggregated = cube.aggregated_by(aggrs, MEAN)
        data = GriddedData(aggregated, **self.metadata)
        data.metadata['ts_type'] = to_ts_type
        data.check_dimcoords_tseries()
        return data     
    
    def downscale_time(self, to_ts_type='monthly'):
        msg = DeprecationWarning('This method is deprecated. Please use new '
                                 'name resample_time')
        print_log.warning(msg)
        return self.resample_time(to_ts_type)
    
    def add_aggregator(self, aggr_name):
        raise NotImplementedError

    def calc_area_weights(self):
        """Calculate area weights for grid"""
        if not self.has_latlon_dims:
            raise DataDimensionError('Data does not have latitude and longitude '
                                     'dimensions. This is required for '
                                     'computation of area weights.')
        self._check_lonlat_bounds()
        self._area_weights = area_weights(self.grid)
        return self.area_weights
                
    def crop(self, lon_range=None, lat_range=None, 
             time_range=None, region=None):
        """High level function that applies cropping along multiple axes
        
        Note
        ----
            1. For cropping of longitudes and latitudes, the method 
            :func:`iris.cube.Cube.intersection` is used since it automatically 
            accepts and understands longitude input based on definition 
            0 <= lon <= 360 as well as for -180 <= lon <= 180
            2. Time extraction may be provided directly as index or in form of
            :class:`pandas.Timestamp` objects. 
            
        Parameters
        ----------
        lon_range : :obj:`tuple`, optional
            2-element tuple containing longitude range for cropping. If None, 
            the longitude axis remains unchanged. 
            Example input to crop around meridian: `lon_range=(-30, 30)`
        lat_range : :obj:`tuple`, optional
            2-element tuple containing latitude range for cropping. If None, 
            the latitude axis remains unchanged
        time_range : :obj:`tuple`, optional
            2-element tuple containing time range for cropping. Allowed data
            types for specifying the times are 
            
                1. a combination of 2 :class:`pandas.Timestamp` instances or 
                2. a combination of two strings that can be directly converted\
                into :class:`pandas.Timestamp` instances (e.g.\
                `time_range=("2010-1-1", "2012-1-1")`) or
                3. directly a combination of indices (:obj:`int`). 
            
            If None, the time axis remains unchanged.
        region : :obj:`str` or :obj:`Region`, optional
            string ID of pyaerocom default region or directly an instance of 
            the :class:`Region` object. May be used instead of 
            ``lon_range`` and ``lat_range``, if these are unspecified.
            
        Returns
        -------
        GriddedData
            new data object containing cropped grid
        """
        suppl = {}
        suppl.update(self.metadata)
        if region is not None:
            if isinstance(region, str):
                try:
                    region = Region(region)
                except Exception as e:
                    logger.warning("Failed to access longitude / latitude range "
                                   "using region ID {}. Error msg: {}".format(
                                           region, repr(e)))
            if not isinstance(region, Region):
                raise ValueError("Invalid input for region")
            suppl["region"] = region.name
            lon_range, lat_range = region.lon_range, region.lat_range
        if lon_range is not None and lat_range is not None:
            data = self.grid.intersection(longitude=lon_range, 
                                          latitude=lat_range)
        elif lon_range is not None and lat_range is None:
            data = self.grid.intersection(longitude=lon_range)
        elif lon_range is None and lat_range is not None:
            data = self.grid.intersection(latitude=lat_range)
        else:
            data = self.grid
        if not data:
            raise DataExtractionError("Failed to apply spatial cropping...")
        if time_range is None:
            return GriddedData(data, **suppl)
        else:
            if all(isinstance(x, str) for x in time_range):
                time_range = (Timestamp(time_range[0]),
                              Timestamp(time_range[1]))
            if all(isinstance(x, Timestamp) for x in time_range):
                logger.info("Cropping along time axis based on Timestamps")
                time_constraint = get_time_rng_constraint(*time_range)
                data = data.extract(time_constraint)
            elif all(isinstance(x, int) for x in time_range):
                logger.info("Cropping along time axis based on indices")
                data = data[time_range[0]:time_range[1]]
            if not data:
                raise DataExtractionError("Failed to apply temporal cropping")
        return GriddedData(data, **suppl)
        
    
    def area_weighted_mean(self):
        """Get area weighted mean"""
        ws = self.area_weights
        return self.collapsed(coords=["longitude", "latitude"], 
                              aggregator=MEAN, 
                              weights=ws).grid.data
        
    def get_area_weighted_timeseries(self, region=None):
        """Helper method to extract area weighted mean timeseries
        
        Parameters
        ----------
        region
            optional, name of AeroCom default region for which the mean is to 
            be calculated (e.g. EUROPE)
        
        Returns
        -------
        StationData
            station data containing area weighted mean
        """
        try:
            self.check_dimcoords_tseries()
        except DimensionOrderError:
            self.reorder_dimensions_tseries()
        if self.ndim != 3:
            raise NotImplementedError('Area weighted mean can only computed '
                                      'for data containing latitude and '
                                      'longitude data')
        stat = StationData()
        stat.station_name = self.data_id
        if region is not None:
            d = self.crop(region=region)
            stat['region'] = region
        else:
            d = self
        vals = d.area_weighted_mean()
         
        stat[self.var_name] = Series(vals, d.time_stamps())
        return stat
        
    # redefined methods from iris.Cube class. This includes all Cube 
    # processing methods that exist in the Cube class and that work on the 
    # Cube and return a Cube instance. These may be expanded (e.g. for 
    # instance what they accept as input
    def aerocom_filename(self, at_stations=False):
        """Filename of data following Aerocom 3 conventions
        
        Parameters
        ----------
        at_stations : str
            if True, then AtStations string will be included in filename
    
        Returns
        -------
        str
            generated file name based on what is in this object
        """
        from pyaerocom.io import FileConventionRead
        f = self.from_files[0]
        fconv = FileConventionRead().from_file(f)
        base_info = fconv.get_info_from_file(f)
    
        vert_code = base_info['vert_code']
        if vert_code is None:
            vert_code = 'UNDEFINED'
        if at_stations:
            vert_code += 'AtStations'
            
        name = [fconv.name, self.name, self.var_name, vert_code,
                str(pd.Timestamp(self.start).year), self.ts_type]
        return '_'.format(fconv.file_sep).join(name) + '.nc'
    
    def compute_at_stations_file(self, latitudes=None, longitudes=None,
                                 out_dir=None, savename=None,
                                 obs_data=None):
        """Creates and saves new netcdf file at input lat / lon coordinates
        
        This method can be used to reduce the size of too large grid files.
        It reduces the lon / lat dimensionality corresponding to the locations
        of the input lat / lon coordinates.

        """
        from pyaerocom import UngriddedData, print_log
        print_log.info('Computing AtStations file. This may take a while')
        if isinstance(obs_data, UngriddedData):
            longitudes = obs_data.longitude
            latitudes = obs_data.latitude
        
        if not len(longitudes) == len(latitudes):
            raise ValueError('Longitude and latitude arrays need to have the '
                             'same length (since they are supposed to belong) '
                             'to station_coordinates')
        if out_dir is None:
            out_dir = const.CACHEDIR
        if savename is None:
            savename = self.aerocom_filename(at_stations=True)
        lons = self.longitude.points
        lats = self.latitude.points
        
        lon_idx = []
        lat_idx = []
        
        for lat, lon in zip(latitudes, longitudes):
            lon_idx.append(closest_index(lons, lon))
            lat_idx.append(closest_index(lats, lat))
        
        lon_idx = sorted(dict.fromkeys(lon_idx))
        lat_idx = sorted(dict.fromkeys(lat_idx))
        try:
            self.check_dimcoords_tseries()
        except DimensionOrderError:
            self.reorder_dimensions_tseries()
          
        subset = self[:, lat_idx][:,:,lon_idx]
        # make sure everything went well with the dimensions
        subset.check_dimcoords_tseries()
        path = subset.to_netcdf(out_dir, savename)
        print_log.info('Finished computing AtStations file.')
        return path
        
    def to_netcdf(self, out_dir, savename=None):
        """Save as netcdf file
        
        Paraemeters
        -----------
        out_dir : str
            output direcory (must exist)
        savename : :obj:`str`, optional
            name of file. If None, :func:`aerocom_filename` is used
            
        Returns
        -------
        str
            file path
        """
        if savename is None:
            savename = self.aerocom_filename()
        fp = os.path.join(out_dir, savename)
        iris.save(self.grid, fp)
        return fp
        
    def interpolate(self, sample_points=None, scheme="nearest", 
                    collapse_scalar=True, **coords):
        """Interpolate cube at certain discrete points
        
        Reimplementation of method :func:`iris.cube.Cube.interpolate`, for 
        details `see here <http://scitools.org.uk/iris/docs/v1.10.0/iris/iris/
        cube.html#iris.cube.Cube.interpolate>`__
        
        Note
        ----
        The input coordinates may also be provided using the input arg **coords
        which provides a more intuitive option (e.g. input
        ``(sample_points=[("longitude", [10, 20]), ("latitude", [1, 2])])`` 
        is the same as input
        ``(longitude=[10, 20], latitude=[1,2])``
        
        
        Parameters
        ----------
        sample_points : list
            sequence of coordinate pairs over which to interpolate
        scheme : str or iris interpolator object
            interpolation scheme, pyaerocom default is Nearest. If input is 
            string, it is converted into the corresponding iris Interpolator 
            object, see :func:`str_to_iris` for valid strings
        collapse_scalar : bool
            Whether to collapse the dimension of scalar sample points in the
            resulting cube. Default is True.
        **coords
            additional keyword args that may be used to provide the interpolation
            coordinates in an easier way than using the ``Cube`` argument
            :arg:`sample_points``. May also be a combination of both.
         
        Returns
        -------
        GriddedData
            new data object containing interpolated data
            
        Examples
        --------
        
            >>> from pyaerocom import GriddedData
            >>> data = GriddedData()
            >>> data._init_testdata_default()
            >>> itp = data.interpolate([("longitude", (10)),
            ...                         ("latitude" , (35))])
            >>> print(itp.shape)
            (365, 1, 1)
        """
        if self._size_GB > self._MAX_SIZE_GB:
            raise MemoryError('Data is too large (grid size: {}, file: {} GB) '
                              'for interpolation (which requires loading data '
                              'into memory)'.format(self.shape, self._size_GB))
        if isinstance(scheme, str):
            scheme = str_to_iris(scheme)
        if not sample_points:
            sample_points = []
        sample_points.extend(list(coords.items()))
        print_log.info('Interpolating data of shape {}. This may take a while.'
                       .format(self.shape))
        try:
            itp_cube = self.grid.interpolate(sample_points, scheme, 
                                             collapse_scalar)
        except MemoryError:
            raise MemoryError("Interpolation failed since grid of interpolated "
                              "Cube is too large")
        print_log.info('Successfully interpolated cube')
        return GriddedData(itp_cube, **self.metadata)
    
    def regrid(self, other, scheme='areaweighted', **kwargs):
        """Regrid this grid to grid resolution of other grid
        
        Parameters
        ----------
        other : GriddedData
            other data object
        scheme : str
            regridding scheme (e.g. linear, neirest, areaweighted)
            
        Returns
        -------
        GriddedData 
            regridded data object (new instance, this object remains unchanged)
        """
        if not isinstance(other, GriddedData):
            other = GriddedData(other)
        if isinstance(scheme, str):
            scheme = str_to_iris(scheme, **kwargs)
            
        self._check_lonlat_bounds()
        other._check_lonlat_bounds()
        data_rg = self.grid.regrid(other.grid, scheme)
        suppl = od(**self.metadata)
        suppl['regridded'] = True
        return GriddedData(data_rg, **suppl)        
    
    def collapsed(self, coords, aggregator, **kwargs):
        """Collapse cube
        
        Reimplementation of method :func:`iris.cube.Cube.collapsed`, for 
        details `see here <http://scitools.org.uk/iris/docs/latest/iris/iris/
        cube.html#iris.cube.Cube.collapsed>`__
        
        Parameters
        ----------
        coords : str or list
            string IDs of coordinate(s) that are to be collapsed (e.g. 
            ``["longitude", "latitude"]``)
        aggregator : str or Aggregator or WeightedAggretor
            the aggregator used. If input is string, it is converted into the
            corresponding iris Aggregator object, see 
            :func:`str_to_iris` for valid strings
        **kwargs 
            additional keyword args (e.g. ``weights``)
        
        Returns
        -------
        GriddedData
            collapsed data object
        """
        if isinstance(aggregator, str):
            aggregator = str_to_iris(aggregator)
        collapsed = self.grid.collapsed(coords, aggregator, **kwargs)
        return GriddedData(collapsed, **self.cube.attributes)
    
    def extract(self, constraint):
        """Extract subset
        
        Parameters
        ----------
        constraint : iris.Constraint
            constraint that is to be applied
            
        Returns
        -------
        GriddedData
            new data object containing cropped data
        """
        data_crop = self.grid.extract(constraint)
        if not data_crop:
            raise DataExtractionError("Failed to extract subset")
        
        return GriddedData(data_crop, **self.metadata)
    
    def intersection(self, *args, **kwargs):
        """Ectract subset using :func:`iris.cube.Cube.intersection` 
        
        See `here for details <http://scitools.org.uk/iris/docs/v1.9.0/html/
        iris/iris/cube.html#iris.cube.Cube.intersection>`__
        related to method and input parameters.
        
        Note
        ----
        Only works if underlying grid data type is :class:`iris.cube.Cube`
        
        Parameters
        ----------
        *args
            non-keyword args
        **kwargs
            keyword args
        
        Returns
        -------
        GriddedData
            new data object containing cropped data
        """
        data_crop = self.grid.intersection(*args, **kwargs)
        
        return GriddedData(data_crop, **self.metadata)
    
    
    def quickplot_map(self, time_idx=0, xlim=(-180, 180), ylim=(-90, 90),
                      add_mean=True, **kwargs):
        """Make a quick plot onto a map
        
        Parameters
        ----------
        time_idx : int
            index in time to be plotted
        xlim : tuple
            2-element tuple specifying plotted longitude range
        ylim : tuple
            2-element tuple specifying plotted latitude range
        add_mean : bool
            if True, the mean value over the region and period is inserted
        **kwargs
            additional keyword arguments passed to 
            :func:`pyaerocom.quickplot.plot_map`
        
        Returns
        -------
        fig
            matplotlib figure instance containing plot
        """
        if not 'latitude' in self.dimcoord_names:
            raise DataDimensionError('Missing latitude dimension...')
        elif not 'longitude' in self.dimcoord_names:
            raise DataDimensionError('Missing longitude dimension...')
        tstr = ''
        if 'time' in self.dimcoord_names:
            if not self.ndim == 3:
                raise DataDimensionError('Invalid number of dimensions: {}. '
                                         'Expected 3.'.format(self.ndim))
            if not isinstance(time_idx, int):
                try:
                    t = to_pandas_timestamp(time_idx).to_datetime64()
                    time_idx = np.argmin(abs(self.time_stamps() - t))
                except:
                    raise ValueError('Failed to interpret input time stamp')
            
            data = self[time_idx]
            try:
                t = cftime_to_datetime64(self.time[time_idx])[0]
                tstr = datetime2str(t, self.ts_type)
            except:
                try:
                    tstr = datetime2str(self.time_stamps()[time_idx], 
                                        self.ts_type)
                except:
                    print_log.warning('Failed to retrieve ts_type in '
                                      'GriddedData {}'.format(repr(self)))
        else:
            if not self.ndim == 2:
                raise DataDimensionError('Invalid number of dimensions: {}. '
                                         'Expected 2.'.format(self.ndim))
            data = self
        
        from pyaerocom.plot.mapping import plot_griddeddata_on_map 
        
        lons = self.longitude.contiguous_bounds()
        lats = self.latitude.contiguous_bounds()
        
        fig = plot_griddeddata_on_map(data=data.grid.data, lons=lons, 
                                      lats=lats, 
                                      var_name=self.var_name, 
                                      unit=self.units,
                                      xlim=xlim, ylim=ylim, 
                                      **kwargs)
        
        fig.axes[0].set_title("{} ({}, {})".format(self.data_id, 
                              self.var_name, tstr))
        if add_mean:
            from pyaerocom.plot.config import COLOR_THEME
            
            ax = fig.axes[0]
            try:
                mean = data.area_weighted_mean()
                mustr = 'Mean={:.2f}'.format(mean)
                u = str(self.units)
                if not u=='1':
                    mustr += ' [{}]'.format(u)
                ax.text(0.02, 0.02, mustr,
                        color=COLOR_THEME.color_map_text, 
                        transform=ax.transAxes,
                        fontsize=22,
                        bbox=dict(facecolor='#ffffff', edgecolor='none', 
                                  alpha=0.65))
            except Exception as e:
                print_log.warning('Failed to compute / add area weighted mean. '
                                  'Reason: {}'.format(repr(e)))
            
        return fig
    
    def min(self):
        """Minimum value"""
        #make sure data is in memory
        data = self.grid.data
        if self.is_masked:
            return data.data[~data.mask].min()
        return data.min()
        
    def max(self):
        """Maximum value"""
        #make sure data is in memory
        data = self.grid.data
        if self.is_masked:
            return data.data[~data.mask].max()
        return data.max()
    
    def mean(self):
        """Mean value of data array
        
        Note
        ----
        Corresponds to numerical mean of underlying N-dimensional numpy array.
        Does not consider area-weights or any other advanced averaging.
        """
        #make sure data is in memory
        data = self.grid.data
        if self.is_masked:
            return data.data[~data.mask].mean()
        return data.mean()
    
    def std(self):
        """Standard deviation of values"""
        #make sure data is in memory
        data = self.grid.data
        if self.is_masked:
            return data.data[~data.mask].std()
        return data.std()
    
    def short_str(self):
        """Short string representation"""
        return "ID: {}, Var: {}".format(self.data_id, self.var_name)
    
    def _check_lonlat_bounds(self):
        """Check if longitude and latitude bounds are set and if not, guess"""
        if self.longitude.bounds is None:
            self.longitude.guess_bounds()
        if self.latitude.bounds is None:
            self.latitude.guess_bounds()

    def _init_testdata_default(self):
        """Helper method that loads ECMWF_OSUITE test data"""
        from pyaerocom.io.testfiles import get
        self.load_input(get()["models"]["ecmwf_osuite"], var_name="od550aer")
        return self
    
    @property
    def _size_GB(self):
        """Size of original data files from which this object is created
        
        These method is intended to be used for operations that actually 
        require the *realisation* of the (lazy loaded) data. 
        """
        return sum([os.path.getsize(f) for f in self.from_files]) / 10**9
    
    def __getattr__(self, attr):
        return self[attr]
        
    def _check_coordinate_access(self, val):
        if self._coord_standard_names is None:
            self._update_coord_info()
        if val in self._coord_standard_names:
            return {'standard_name' : val}
        elif val in self._coord_var_names:
            return {'var_name' : val}
        elif val in self._coord_long_names:
            return {'long_name' : val}
        raise CoordinateError('Could not associate one of the coordinates with '
                              'input string {}'.format(val))
       
    def copy(self):
        """Copy this data object"""
        return GriddedData(self.cube.copy())
     
    def delete_aux_vars(self):
        """Delete auxiliary variables and iris AuxFactories"""
        c = self.cube
        for aux_fac in c.aux_factories:
    
            c.remove_aux_factory(aux_fac)
    
        for coord in c.coords():
            if isinstance(coord, iris.coords.AuxCoord):
                c.remove_coord(coord.name())
        
    def search_other(self, var_name, require_same_shape=True):
        """Searches data for another variable"""
        if require_same_shape and self.concatenated or self.computed:
            raise NotImplementedError('Coming soon...')
        for file in self.from_files:
            try:
                from pyaerocom.io.iris_io import load_cube_custom
                cube = load_cube_custom(file, var_name=var_name,
                                        perform__fmt_checks=False)
                return GriddedData(cube, from_files=file)
            except:
                pass
        if var_name in self.reader.vars_provided:
            return self.reader.read_var(var_name,
                                        start=self.start,
                                        stop=self.stop,
                                        ts_type=self.ts_type,
                                        flex_ts_type=True)
        raise VariableNotFoundError('Could not find variable {}'.format(var_name))
    
    def update_meta(self, **kwargs):
        """Update metadata dictionary"""
        for key, val in kwargs.items():
            self._grid.attributes[key] = val
            
    def delete_all_coords(self, inplace=True):
        """Deletes all coordinates (dimension + auxiliary) in this object"""
        if inplace:
            obj = self
        else:
            obj = self.copy()
        delete_all_coords_cube(obj.cube, inplace=True)
        return obj
        
    def copy_coords(self, other, inplace=True):
        """Copy all coordinates from other data object
        
        Requires the underlying data to be the same shape. 
        
        Warning
        --------
        This operation will delete all existing coordinates and auxiliary 
        coordinates and will then copy the ones from the input data object.
        No checks of any kind will be performed
        
        Parameters
        ----------
        other : GriddedData or Cube
            other data object (needs to be same shape as this object)
        
        Returns
        -------
        GriddedData
            data object containing coordinates from other object
        """
        if inplace:
            obj = self
        else:
            obj = self.copy()
        if isinstance(other, iris.cube.Cube):
            other = GriddedData(other)
        if not other.shape == obj.shape:
            raise DataDimensionError('Cannot copy coordinates: shape mismatch')
        copy_coords_cube(to_cube=obj.cube, from_cube=other.cube, inplace=True)
        return obj

    def _update_coord_info(self):
        n, vn, sn, ln = [], [], [], []
        for c in self.grid.coords():
            n.append(c.name())
            vn.append(c.var_name)
            sn.append(c.standard_name)
            ln.append(c.long_name)
        self._coord_names = n
        self._coord_var_names = vn
        self._coord_standard_names = sn
        self._coord_long_names = ln
        
    def __getitem__(self, indices_or_attr):
        """x.__getitem__(y) <==> x[y]"""
   
        if isinstance(indices_or_attr, str):
            if indices_or_attr in self.__dict__:
                return self.__dict__[indices_or_attr]
            try:
                which = self._check_coordinate_access(indices_or_attr)
                return self.grid.coord(**which)
            except:
                raise AttributeError("GriddedData object has no "
                                     "attribute {}"
                                     .format(indices_or_attr))
            
        sub = self.grid.__getitem__(indices_or_attr)
        return GriddedData(sub, **self.metadata)
    
    def __contains__(self, val):
        """Check if variable or coordinate matchs input string"""
        return val is self.name or val in self.coord_names
     
    def __dir__(self):
        return self.coord_names + super().__dir__()
    
    def __str__(self):
        """For now, use string representation of underlying data"""
        return ("pyaerocom.GriddedData: %s\nGrid data: %s"
                %(self.name, self.grid.__str__()))
    
    def __repr__(self):
        """For now, use representation of underlying data"""
        return "pyaerocom.GriddedData\nGrid data: %s" %self.grid.__repr__()
    
    def __add__(self, other):
        raise NotImplementedError('Coming soon')
        
    #sorted out
    def _to_timeseries_iter_coords_2D(self, sample_points, scheme, 
                                      collapse_scalar):
        """Extract time-series for provided input coordinates (lon, lat)
        
        This method extracts the time-series at all input coordinates by 
        iterating over the coordinate locations, cropping the grid around the 
        coordinate and then interpolating it using
        the provided interpolation scheme.
        
        This method may be faster for a small number of coordinates (compared 
        to :func:`to_timeseries`). It may also be the better choice in case the
        number of coordinates is too large in which case :func:`to_time_series`
        may fail due to a MemoryError (i.e. the case where the final 
        interpolated object is too large to fit into memory).
        
        Parameters
        ----------
        sample_points : list
            coordinates (e.g. lon / lat) at which time series is supposed to be
            retrieved
        scheme : str or iris interpolator object
            interpolation scheme (for details, see :func:`interpolate`)
        collapse_scalar : bool
            see :func:`interpolate`
        **coords
            additional keyword args that may be used to provide the interpolation
            coordinates (for details, see :func:`interpolate`)

        Returns
        -------
        list
            list of result dictionaries for each coordinate. Dictionary keys
            are: ``latitude, longitude, altitude, var_name``
        """
        raise NotImplementedError
        if not scheme=="nearest":
            raise NotImplementedError
        self.check_dimcoords_tseries()
        
        lats, lons = None, None
        for val in sample_points:
            name, vals = val[0], val[1]
            if name == 'latitude':
                lats = vals
            elif name == 'longitude':
                lons = vals
        
        var = self.var_name
        times = self.time_stamps()
        grid_lats = self.latitude.points
        grid_lons = self.longitude.points
        result = []
        totnum = len(lats)
        for i, lat in enumerate(lats):
            if i%10 == 0:
                print('At coord {} of {}'.format(i+1, totnum))
            lon = lons[i]
            
            lat_idx = np.argmin(np.abs(grid_lats - lat))
            lon_idx = np.argmin(np.abs(grid_lons - lon))
            
            
            #: TODO review indexing [:,:] style vs. extract method vs. lazy data
            C = iris.Constraint(latitude=grid_lats[lat_idx],
                               longitude=grid_lons[lon_idx])
            
            sub = self.extract(C)
            
            vals = sub.grid.data
            #sub = self.grid[:, lat_idx, lon_idx]
            # first slice, then access data
            data = Series(vals, index=times)
            result.append({'latitude'   :   lat,
                           'longitude'  :   lon,
                           'name'       :   self.name, 
                            var         :   data})
        return result

    ### Deprecated (but still supported) stuff
    @property
    def unit(self):
        """Unit of data"""
        const.print_log.warn(DeprecationWarning('Attr. unit is deprecated, '
                                                'please use units instead'))
        return self.grid.units

    @unit.setter
    def unit(self, val):
        const.print_log.warn(DeprecationWarning('Attr. unit is deprecated, '
                                                'please use units instead'))
        self.grid.units = val

if __name__=='__main__':
    import matplotlib.pyplot as plt
    import pyaerocom as pya
    plt.close("all")
    
    
    data = pya.io.ReadGridded('ECMWF_CAMS_REAN').read_var('od550aer')
    print('Here')
    ts = data.get_area_weighted_timeseries()
    

# =============================================================================
#     data.downscale_time('monthly')
#     
#     t1 = data.to_time_series(longitude=[30], latitude=[40],
#                              vert_scheme='max')
#     
#     t2 = data.to_time_series(longitude=[30], latitude=[40],
#                              vert_scheme='surface')
#     
#     t3 = data.to_time_series(longitude=[30], latitude=[40],
#                              vert_scheme='mean')
#     
#     ax = t1.plot_timeseries('ec532aer')
# 
# =============================================================================



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
from pyaerocom.tstype import TsType

def test_basic_operators():
    monthly = TsType('monthly')
    yearly = TsType('yearly')
    daily = TsType('daily')
    
    assert monthly < daily
    assert monthly <= daily
    assert monthly != daily
    assert yearly < daily
    assert not (yearly == daily)
    assert monthly > yearly
    assert monthly >= yearly
    
def test_basic_operators_pandas():
    monthly = TsType('MS')
    yearly = TsType('AS')
    daily = TsType('D')
    
    assert monthly < daily
    assert monthly <= daily
    assert monthly != daily
    assert yearly < daily
    assert not (yearly == daily)
    assert monthly > yearly
    assert monthly >= yearly

if __name__=="__main__":
    test_basic_operators()
    test_basic_operators_pandas()    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 15 14:00:44 2019

ToDo
----
- the configuration classes could inherit from a base class or could be more unified

"""
import os, glob, shutil
import simplejson
from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict, sort_dict_by_name

def delete_experiment_data_evaluation_iface(base_dir, proj_id, exp_id):
    """Delete all data associated with a certain experiment
        
    Parameters
    ----------
    base_dir : str, optional
        basic output direcory (containg subdirs of all projects)
    proj_name : str, optional
        name of project, if None, then this project is used
    exp_name : str, optional
        name experiment, if None, then this project is used
    """
    
    p = os.path.join(base_dir, proj_id, exp_id)
    if not os.path.exists(p):
        raise NameError('No such data directory found: {}'.format(p))
    
    shutil.rmtree(p)
    
def get_all_config_files_evaluation_iface(config_dir):
    """
    
    Note
    ----
    This code only checks json configuration files, not .py module files
    containing configuration.
    
    Parameters
    ----------
    config_dir : str
        directory containing config json files for AerocomEvaluation interface
        
    Returns
    -------
    dict
        nested dictionary containing file paths of all config files that were
        detected, where first level of dict id `proj_id` and second level 
        is `exp_id`.
    """
    
    results = {}
    
    for file in glob.glob('{}/*.json'.format(config_dir)):
        spl = os.path.basename(file).split('_')
        if not len(spl) == 3 or not spl[0] =='cfg':
            raise NameError('Invalid config file name ', file)
        proj, exp = spl[1], spl[2].split('.')[0]
        if not proj in results:
            results[proj] = {}
        results[proj][exp] = file
    return results

def get_all_config_files_trends_iface(config_dir):
    """Get all configuration files of trends interface
    
    Note
    ----
    This code only checks json configuration files, not .py module files
    containing configuration.
    
    Parameters
    ----------
    config_dir : str
        directory containing config json files for AerosolTrends interface
        
    Returns
    -------
   dictionary
        keys are configuration names, values are corresponding file paths
    """
    results = {}
    for file in  glob.glob('{}/cfg_*.json'.format(config_dir)):
        spl = os.path.basename(file).split('_')
        if not len(spl) == 2 or not spl[0] =='cfg':
            raise NameError('Invalid config file name ', file)
        results[spl[1].split('.')[0]] = file
    return results
      
class ObsConfigEval(BrowseDict):
    """Observation configuration for evaluation (dictionary)
    
    Note
    ----
    Only :attr:`obs_id` and `obs_vars` are mandatory, the rest is optional.
    
    Attributes
    ----------
    obs_id : str
        ID of observation network in AeroCom database 
        (e.g. 'AeronetSunV3Lev2.daily')
    obs_vars : list
        list of pyaerocom variable names that are supposed to be analysed
        (e.g. ['od550aer', 'ang4487aer'])
    obs_ts_type_read : :obj:`str` or :obj:`dict`, optional
        may be specified to explicitly define the reading frequency of the 
        observation data (so far, this does only apply to gridded obsdata such
        as satellites). For ungridded reading, the frequency may be specified
        via :attr:`obs_id`, where applicable (e.g. AeronetSunV3Lev2.daily).
        Can be specified variable specific in form of dictionary.
    obs_vert_type : :obj:`str` or :obj:`dict`, optional
        Aerocom vertical code encoded in the model filenames (only AeroCom 3 
        and later). Specifies which model file should be read in case there are
        multiple options (e.g. surface level data can be read from a 
        *Surface*.nc file as well as from a *ModelLevel*.nc file). If input is 
        string (e.g. 'Surface'), then the corresponding vertical type code is 
        used for reading of all variables that are colocated (i.e. that are 
        specified in :attr:`obs_vars`). Else (if input is dictionary, e.g. 
        `obs_vert_type=dict(od550aer='Column', ec550aer='ModelLevel')`), 
        information is extracted variable specific, for those who are defined
        in the dictionary, for all others, `None` is used.
    read_opts_ungridded : :obj:`dict`, optional
        dictionary that specifies reading constraints for ungridded reading
        (c.g. :class:`pyaerocom.io.ReadUngridded`).
    """
    SUPPORTED_VERT_CODES = ['Column', 'Profile', 'Surface']
    ALT_NAMES_VERT_CODES = dict(ModelLevel = 'Profile')
    def __init__(self, **kwargs):
        
        self.obs_id = None
        self.obs_vars = None
        self.obs_ts_type_read = None
        self.obs_vert_type = None
        
        self.read_opts_ungridded = None
        
        self.update(**kwargs)
        self.check_cfg()
    
    def check_cfg(self):
        """Check that minimum required attributes are set and okay"""
        if not isinstance(self.obs_id, str):
            raise ValueError('Invalid value for obs_id: {}. Need str.'
                             .format(self.obs_id))
        if isinstance(self.obs_vars, str):
            self.obs_vars = [self.obs_vars]
        elif not isinstance(self.obs_vars, list):
            raise ValueError('Invalid input for obs_vars. Need list or str, '
                             'got: {}'.format(self.obs_vars))
        ovt = self.obs_vert_type
        if ovt is None:
            raise ValueError('obs_vert_type is not defined. Please specify '
                             'using either of the available codes: {}. '
                             'It may be specified for all variables (as string) '
                             'or per variable using a dict'
                             .format(self.SUPPORTED_VERT_CODES))
        elif (isinstance(ovt, str) and not ovt in self.SUPPORTED_VERT_CODES):
            self.obs_vert_type = self._check_ovt(ovt)
        elif isinstance(self.obs_vert_type, dict):
            for var_name, val in self.obs_vert_type.items():
                if not val in self.SUPPORTED_VERT_CODES:
                    raise ValueError('Invalid value for obs_vert_type: {} '
                                     '(variable {}). Supported codes are {}.'
                                     .format(self.obs_vert_type,
                                             var_name,
                                             self.SUPPORTED_VERT_CODES))
    def _check_ovt(self, ovt):
        """Check if obs_vert_type string is valid alias
        
        Parameters
        ----------
        ovt : str
            obs_vert_type string
        
        Returns
        -------
        str
            valid obs_vert_type
        
        Raises
        ------
        ValueError
            if `ovt` is invalid
        """
        if ovt in self.ALT_NAMES_VERT_CODES:
            _ovt = self.ALT_NAMES_VERT_CODES[ovt]
            const.print_log.warning('Please use {} for obs_vert_code '
                                        'and not {}'.format(_ovt, ovt))
            return _ovt
        valid = self.SUPPORTED_VERT_CODES + list(self.ALT_NAMES_VERT_CODES.keys())
        raise ValueError('Invalid value for obs_vert_type: {}. '
                         'Supported codes are {}.'
                         .format(self.obs_vert_type,
                                 valid))
            
        
class ModelConfigEval(BrowseDict):
    """Modeln configuration for evaluation (dictionary)
    
    Note
    ----
    Only :attr:`model_id` is mandatory, the rest is optional.
    
    Attributes
    ----------
    model_id : str
        ID of model run in AeroCom database (e.g. 'ECMWF_CAMS_REAN')
    model_ts_type_read : :obj:`str` or :obj:`dict`, optional
        may be specified to explicitly define the reading frequency of the 
        model data. Not to be confused with :attr:`ts_type`, which specifies 
        the frequency used for colocation. Can be specified variable specific 
        by providing a dictionary.
    model_use_vars : :obj:`dict`, optional
        dictionary that specifies mapping of model variables. Keys are 
        observation variables, values are the corresponding model variables 
        (e.g. model_use_vars=dict(od550aer='od550csaer'))
    model_read_aux : :obj:`dict`, optional
        may be used to specify additional computation methods of variables from
        models. Keys are obs variables, values are dictionaries with keys 
        `vars_required` (list of required variables for computation of var 
        and `fun` (method that takes list of read data objects and computes
        and returns var)    
    """
    def __init__(self, model_id, **kwargs):
        self.model_id = model_id
        self.model_ts_type_read = None
        self.model_use_vars = {}
        self.model_read_aux = {}
        
        self.update(**kwargs)
        self.check_cfg()
    
    def check_cfg(self):
        """Check that minimum required attributes are set and okay"""
        if not isinstance(self.model_id, str):
            raise ValueError('Invalid input for model_id {}. Need str.'
                             .format(self.model_id))
            
def update_menu_evaluation_iface(config, ignore_experiments=None):
    """Update menu for Aerocom Evaluation interface
    
    The menu.json file is created based on the available json map files in the
    map directory of an experiment.
    
    Parameters
    ----------
    config : AerocomEvaluation
        instance of config class that has all relevant paths specified
    ignore_experiments : list, optional
        list containing experiment IDs that may be in the current menu.json 
        file and that are supposed to be removed from it.
    """
    def var_dummy():
        """Helper that creates empty dict for variable info"""
        return {'type'      :   '',
                'cat'       :   '',
                'name'      :   '',
                'longname'  :   '',
                'obs'       :   {}}
        
    new = {}
    if ignore_experiments is None:
        ignore_experiments = []
    try:
        tab = config.get_web_overview_table()
    except FileNotFoundError:
        import pandas as pd
        tab = pd.DataFrame()
    
    for index, info in tab.iterrows():
        obs_var, obs_name, vert_code, mod_name, mod_var = info
        if not obs_var in new:
            const.print_log.info('Adding new observation variable: {}'
                                 .format(obs_var))
            new[obs_var] = d = var_dummy()
            name, tp, cat = config.get_obsvar_name_and_type(obs_var)
    
            d['name'] = name
            d['type'] = tp
            d['cat']  =cat
            d['longname'] = const.VARS[obs_var].description
        else:
            d = new[obs_var]
        
        if not obs_name in d['obs']:
            d['obs'][obs_name] = dobs = {}
        else:
            dobs = d['obs'][obs_name]
        
        if not vert_code in dobs:
            dobs[vert_code] = dobs_vert = {}
        else:
            dobs_vert = dobs[vert_code]
        if mod_name in dobs_vert:
            raise Exception
        dobs_vert[mod_name] = {'dir' : mod_name,
                               'id'  : config.model_config[mod_name]['model_id'],
                               'var' : mod_var}
        
    _new = {}
    for var in config.var_order_menu:
        try:
            _new[var] = new[var]
        except:
            const.print_log.info('No variable {} found'.format(var))
    for var, info in new.items():
        if not var in _new:
            _new[var] = info
    new = _new
    new_sorted = {}
    for var, info in new.items():
        new_sorted[var] = info
        sorted_obs = sort_dict_by_name(info['obs'])
        new_sorted[var]['obs'] = sorted_obs
        for obs_name, vert_codes in sorted_obs.items():
            vert_codes_sorted = sort_dict_by_name(vert_codes)
            new_sorted[var]['obs'][obs_name] = vert_codes_sorted
            for vert_code, models in vert_codes_sorted.items():
                models_sorted = sort_dict_by_name(models)
                new_sorted[var]['obs'][obs_name][vert_code] = models_sorted
    new_menu = {}
    basedir = os.path.dirname(config.menu_file)
    available_exps = os.listdir(basedir)
    if os.path.exists(config.menu_file):
        current = read_json(config.menu_file)
        for exp, submenu in current.items():
            if exp in ignore_experiments:
                continue
            elif not exp in available_exps:
                const.print_log.info('Removing outdated experiment {} from '
                                     'menu.json'.format(exp))
                continue
            new_menu[exp] = submenu
    
    if config.exp_id in new_menu:
        const.print_log.warning('Sub menu for experiment {} already exists in '
                                'menu.json and will be overwritten'
                                .format(config.exp_id))
    
    new_menu[config.exp_id] = new_sorted
    
    with open(config.menu_file, 'w+') as f:
        f.write(simplejson.dumps(new_menu, indent=4))
 
def read_json(file_path):
    """Read json file
    
    Parameters
    ----------
    file_path : str
        json file path
    
    Returns
    -------
    dict
        content as dictionary
    """
    with open(file_path, 'r') as f:
        data = simplejson.load(f)
    return data

def write_json(data_dict, file_path, indent=4):
    """Save json file
    
    Parameters
    ----------
    data_dict : dict
        dictionary that can be written to json file
    file_path : str
        output file path
    """
    with open(file_path, 'w+') as f:
        f.write(simplejson.dumps(data_dict, indent=4))

def update_menu_trends_iface(config):
    """Update menu for Aerosol trends interface
    
    The menu.json file is created based on the available json map files in the
    map directory of an experiment.
    
    Parameters
    ----------
    config : AerocomEvaluation
        instance of config class that has all relevant paths specified
    """
    def var_dummy():
        """Helper that creates empty dict for variable info"""
        return {'type'      :   '',
                'cat'       :   '',
                'name'      :   '',
                'longname'  :   '',
                'obs'       :   {}}
        
    new = {}
    
    tab = config.get_web_overview_table()
    
    for index, info in tab.iterrows():
        (obs_name, obs_id, obs_var, 
         vert_code, 
         mod_name, mod_id, mod_var, mod_type, 
         periods) = info
        if not obs_var in new:
            const.print_log.info('Adding new observation variable: {}'
                                 .format(obs_var))
            new[obs_var] = d = var_dummy()
            name, tp, cat = config.get_obsvar_name_and_type(obs_var)
    
            d['name'] = name
            d['type'] = tp
            d['cat'] = cat
            d['longname'] = const.VARS[obs_var].description
        else:
            d = new[obs_var]
        
        if not obs_name in d['obs']:
            d['obs'][obs_name] = dobs = {}
        else:
            dobs = d['obs'][obs_name]
        
        if not vert_code in dobs:
            dobs[vert_code] = dobs_vert = dict(obs_id=obs_id,
                                               period=periods,
                                               modsat= {})
        else:
            dobs_vert = dobs[vert_code]
        if mod_name is not None:
            if mod_name in dobs_vert['modsat']:
                raise KeyError('Model {} already exists in {}. This should '
                               'not occur. please debug...'.format(mod_name, dobs))
            dobs_vert['modsat'][mod_name] = dict(var= mod_var,
                                                 mtype=mod_type,
                                                 mid=mod_id)
        
    _new = {}
    for var in config.var_order_menu:
        try:
            _new[var] = new[var]
        except:
            const.print_log.info('No variable {} found'.format(var))
    for var, info in new.items():
        if not var in _new:
            _new[var] = info
    new = _new
    if config.obs_order_menu_cfg:
        pref_obs_order = list(config.obs_config)
    else:
        pref_obs_order = []
    new_sorted = {}
    for var, info in new.items():
        new_sorted[var] = info
        sorted_obs = sort_dict_by_name(info['obs'],
                                       pref_list=pref_obs_order)
        new_sorted[var]['obs'] = sorted_obs
        for obs_name, vert_codes in sorted_obs.items():
            vert_codes_sorted = sort_dict_by_name(vert_codes)
            new_sorted[var]['obs'][obs_name] = vert_codes_sorted
            for vert_code, models in vert_codes_sorted.items():
                models_sorted = sort_dict_by_name(models)
                new_sorted[var]['obs'][obs_name][vert_code] = models_sorted
    
    with open(config.menu_file, 'w+') as f:
        f.write(simplejson.dumps(new_sorted, indent=4))
        
def make_info_table_evaluation_iface(config):
    from pyaerocom import ColocatedData
    import glob
    menu = config.menu_file
        
        
    SKIP_META = ['data_source', 'var_name', 'lon_range',
                 'lat_range', 'alt_range']
    with open(menu, 'r') as f:
        menu = simplejson.load(f)
    if not config.exp_id in menu:
        raise KeyError('No menu entry available for experiment {}'
                       .format(config.exp_id))
    table = {}
    exp = menu[config.exp_id]
    for obs_var, info in exp.items():
        for obs_name, vert_types in info['obs'].items():
            for vert_type, models in vert_types.items():
                for mname, minfo in models.items():
                    if not mname in table:
                        table[mname] = mi = {}
                        mi['id'] = model_id = minfo['id']
                    else:
                        mi = table[mname]
                        model_id = mi['id']
                        if minfo['id'] != mi['id']:
                            raise KeyError('Unexpected error: conflict in model ID and name')
                    
                    try:
                        mo = mi['obs']
                    except:
                        mi['obs'] = mo = {}
                    if 'var' in minfo:
                        mvar = minfo['var']
                    else:
                        mvar = obs_var
                    if not obs_var in mo:
                        mo[obs_var] = oi = {}
                    else: 
                        oi = mo[obs_var]
                    if obs_name in oi:
                        raise Exception
                    oi[obs_name] = motab = {}
                    motab['model_var'] = mvar
                    motab['obs_id'] = config.get_obs_id(obs_name)
                    files = glob.glob('{}/{}/{}*REF-{}*.nc'
                                      .format(config.coldata_dir, 
                                              model_id, mvar, obs_name))
                    
                    if not len(files) == 1:
                        if len(files) > 1:
                            motab['MULTIFILES'] = len(files)
                        else:
                            motab['NOFILES'] = True
                        continue
                    
                    coldata = ColocatedData(files[0])
                    for k, v in coldata.meta.items():
                        if not k in SKIP_META:
                            if isinstance(v, (list, tuple)):
                                if not len(v) == 2:
                                    raise Exception
                            
                                motab['{}_obs'.format(k)] = str(v[0])
                                motab['{}_mod'.format(k)] = str(v[1])
                            else:
                                motab[k] = str(v)
    return table

if __name__ == '__main__':
    print()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 28 13:26:36 2019

@author: jonasg
"""
import os
import numpy as np
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom import const

class DataSource(BrowseDict):
    """Dict-like object defining a data source
    
    Attributes
    ----------
    data_id
        name (or ID) of dataset (e.g. AeronetSunV3Lev2.daily)
    dataset_name
        name of dataset (e.g. AERONET)
    data_product
        data product (e.g. SDA, Inv, Sun for Aeronet)
    data_version
        version of data (e.g. 3)
    data_level 
        level of data (e.g. 2)
    revision_date
        last revision date of dataset
    ts_type_src 
        sampling frequency as defined in data files (use None if undefined)
    stat_merge_pref_attr : str
        optional, a metadata attribute that is available in data and that
        is used to order the individual stations by relevance in case overlaps
        occur. The associated values of this attribute need to be sortable 
        (e.g. revision_date). This is only relevant in case overlaps occur. 
    """
    _types = dict(dataset_name          =   str,
                  data_product          =   str,
                  data_version          =   float,
                  data_level            =   float,
                  ts_type_src           =   str,
                  stat_merge_pref_attr  =   str, 
                  revision_date         =   np.datetime64,
                  website               =   str)
    
    _ini_file_name = 'data_sources.ini'
    def __init__(self, **info):

        self.data_id = None
        self.dataset_name = None
        self.data_product = None
        self.data_version = None
        self.data_level = None
        self.revision_date = None
        self.website = None
        
        self.ts_type_src = None
        
        self.stat_merge_pref_attr = None
        
        self.update(**info)
        if self.data_id is not None:
            self._parse_source_info_from_ini()
        
    @property
    def data_dir(self):
        """Directory containing data files"""
        from pyaerocom.io.helpers import get_obsnetwork_dir
        return get_obsnetwork_dir(self.data_id)
    
    
    def dataset_str(self):
        s = ''
        if self.dataset_name is not None:
            s += self.dataset_name
            hasv = False
            if self.data_version is not None:
                s += '(v{}'.format(self.data_version)
                hasv = True
            if self.data_level is not None:
                if hasv:
                    s += ', Lev {})'.format(self.data_level)
                else:
                    s += '(Lev {})'.format(self.data_level)
            else:
                s += ')'
        else:
            s += self.data_id
        return s
    
    def load_dataset_info(self):
        """Wrapper for :func:`_parse_source_info_from_ini`"""
        try:
            self._parse_source_info_from_ini()
        except:
            pass
        
    def _parse_source_info_from_ini(self):
        """Parse source info from ini file"""
        try:
            from ConfigParser import ConfigParser
        except:
            from configparser import ConfigParser
        cfg = ConfigParser()
        file = os.path.join(const.DIR_INI_FILES, self._ini_file_name)
        if not os.path.exists(file):
            raise IOError('File {} does not exist'.format(self._ini_file_name))
        cfg.read(file)
        if self.data_id in cfg:
            for k, v in cfg[self.data_id].items():
                if k in self._types:
                    self[k] = self._types[k](v)
                else:
                    self[k] = str(v)
              
class StationMetaData(DataSource):
    """This object defines a standard for station metadata in pyaerocom
    
    Variable names associated with meta data can vary significantly between 
    different conventions (e.g. conventions in modellers community vs. 
    observations community). 
    
    Note
    ----
    - This object is a dictionary and can be easily expanded
    - In many cases, only some of the attributes are relevant
    
    Attributes
    ----------
    filename : str
        name of file (may be full path or only filename)
    station_id : str
        Code or unique ID of station
    station_name :str
        name or ID of a station. Note, that the concept of a station in 
        pyaerocom is not necessarily related to a fixed coordinate. A station
        can also be a satellite, ship, or a human walking around and measuring
        something
    instrument_name : str
        name (or ID) of instrument
    PI : str
        principal investigator
    country : str
        string specifying country (or country ID)
    ts_type : str
        frequency of data (e.g. monthly). Note the difference between 
        :attr:`ts_type_src` of :class:`DataSource`, which specifies the freq.
        of the original files.
    latitude : float
        latitude coordinate
    longitude : float
        longitude coordinate
    altitude : float
        altitude coordinate
    
    """
    def __init__(self, **info):
        
        self.filename = None
        
        self.station_id = None
        self.station_name = None
        self.instrument_name = None
        self.PI = None
        
        self.country = None
        
        self.ts_type = None
        
        self.latitude = np.nan
        self.longitude = np.nan
        self.altitude = np.nan
        
        super(StationMetaData, self).__init__(**info)
                
     
if __name__ == '__main__':
    meta = StationMetaData(data_id = 'AeronetSunV3Lev2.daily',
                           ts_type = 'blaaaa')
    print(meta)
    print(meta.dataset_str())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Settings for I/O of obervation data 

Note
----
Some settings like paths etc can be found in :mod:`pyaerocom.config.py`
"""
#: Wavelength tolerance for observations if data for required wavelength
        #: is not available
OBS_WAVELENGTH_TOL_NM = 10.0
        
#: This boolean can be used to enable / disable the former (i.e. use
#: available wavelengths of variable in a certain range around variable
#: wavelength).
OBS_ALLOW_ALT_WAVELENGTHS = True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
General helper methods for the pyaerocom library.
"""
from cf_units import Unit
from datetime import MINYEAR, datetime, date
import iris
import numpy as np
import pandas as pd
import xarray as xray

from pyaerocom.exceptions import (LongitudeConstraintError, 
                                  DataCoverageError, MetaDataError,
                                  DataDimensionError)
from pyaerocom import logger, const
from pyaerocom.time_config import (GREGORIAN_BASE, TS_TYPE_SECS,
                                   TS_TYPE_TO_PANDAS_FREQ,
                                   PANDAS_RESAMPLE_OFFSETS,
                                   TS_TYPE_DATETIME_CONV,
                                   microsec_units, millisec_units,
                                   sec_units, min_units, hr_units,
                                   day_units)

NUM_KEYS_META = ['longitude', 'latitude', 'altitude']

STR_TO_IRIS = dict(count       = iris.analysis.COUNT,
                   gmean       = iris.analysis.GMEAN,
                   hmean       = iris.analysis.HMEAN,
                   max         = iris.analysis.MAX,
                   mean        = iris.analysis.MEAN,
                   median      = iris.analysis.MEDIAN,
                   sum         = iris.analysis.SUM,
                   nearest     = iris.analysis.Nearest,
                   linear      = iris.analysis.Linear,
                   areaweighted= iris.analysis.AreaWeighted)

def delete_all_coords_cube(cube, inplace=True):
    """Delete all coordinates of an iris cube
    
    Parameters
    ----------
    cube : iris.cube.Cube
        input cube that is supposed to be cleared of coordinates
    inplace : bool
        if True, then the coordinates are deleted in the input object, else in
        a copy of it
        
    Returns
    -------
    iris.cube.Cube
        input cube without coordinates 
    """
    if not inplace:
        cube = cube.copy()
        
    for aux_fac in cube.aux_factories:
        cube.remove_aux_factory(aux_fac)

    for coord in cube.coords():
        cube.remove_coord(coord)
    return cube

def copy_coords_cube(to_cube, from_cube, inplace=True):
    """Copy all coordinates from one cube to another
    
    Requires the underlying data to be the same shape. 
    
    Warning
    --------
    This operation will delete all existing coordinates and auxiliary 
    coordinates and will then copy the ones from the input data object.
    No checks of any kind will be performed
    
    Parameters
    ----------
    to_cube
    other : GriddedData or Cube
        other data object (needs to be same shape as this object)
    
    Returns
    -------
    GriddedData
        data object containing coordinates from other object
    """
    if not all([isinstance(x, iris.cube.Cube) for x in [to_cube, from_cube]]):
        raise ValueError('Invalid input. Need instances of iris.cube.Cube class...')
        
    if not from_cube.shape == to_cube.shape:
        raise DataDimensionError('Cannot copy coordinates: shape mismatch')
    
    to_cube = delete_all_coords_cube(to_cube, inplace)
    
    for i, dim_coord in enumerate(from_cube.dim_coords):
        to_cube.add_dim_coord(dim_coord, i)
    
    for aux_coord, dim in from_cube._aux_coords_and_dims:
        to_cube.add_aux_coord(aux_coord, dim)
        
    for aux_fac in from_cube.aux_factories:
        to_cube.add_aux_factory(aux_fac)
    return to_cube

def infer_time_resolution(time_stamps):
    """Infer time resolution based on input time-stamps
    
    Uses the minimum time difference found in input array between consecutive
    time stamps and based on that finds the corresponding AeroCom resolution
    
    Parameters
    ----------
    time_stamps : pandas.DatetimeIndex
        time stamps 
    
    Ret    
    """
    import pandas as pd
    if not isinstance(time_stamps, pd.DatetimeIndex):
        try:
            time_stamps = pd.DatetimeIndex(time_stamps)
        except:
            raise ValueError('Could not infer time resolution: failed to '
                             'convert input to pandas.DatetimeIndex')
    vals = time_stamps.values
    highest_secs = abs(vals[1:] - vals[:-1]).min().astype('timedelta64[s]').astype(int)
    
    for tp in const.GRID_IO.TS_TYPES:
        if highest_secs <= TS_TYPE_SECS[tp]:
            return tp
    raise ValueError('Could not infer time resolution')

def get_tot_number_of_seconds(ts_type, dtime = None):
    from pyaerocom.tstype import TsType

    ts_tpe = TsType(ts_type)

    if ts_tpe >= TsType('monthly'):
        if dtime is None:
            raise AttributeError('For frequncies larger than or eq. monthly you' +
                                 ' need to provide dtime in order to compute the number of second.')
        else:
            # find seconds from dtime
            # TODO generalize this
            days_in_month = dtime.dt.daysinmonth
            if ts_type == 'monthly':
                monthly_to_sec = days_in_month*24*60*60
            return monthly_to_sec
    else:
        return TS_TYPE_SECS[ts_type]

def get_standard_name(var_name):
    """Converts AeroCom variable name to CF standard name
    
    Also handles alias names for variables, etc. or strings corresponding to
    older conventions (e.g. names containing 3D).
    
    Parameters
    ----------
    var_name : str
        AeroCom variable name
        
    Returns
    -------
    str
        corresponding standard name
    """
    return const.VARS[var_name].standard_name

def get_standard_unit(var_name):
    """Gets standard unit of AeroCom variable
    
    Also handles alias names for variables, etc. or strings corresponding to
    older conventions (e.g. names containing 3D).
    
    Parameters
    ----------
    var_name : str
        AeroCom variable name
        
    Returns
    -------
    str
        corresponding standard unit
    """
    return const.VARS[var_name].unit

def get_lowest_resolution(ts_type, *ts_types):
    """Get the lowest resolution from several ts_type codes
    
    Parameters
    ----------
    ts_type : str
        first ts_type
    *ts_types
        one or more additional ts_type codes
    
    Returns
    -------
    str
        the ts_type that corresponds to the lowest resolution
        
    Raises
    ------
    ValueError
        if one of the input ts_type codes is not supported
    """
    #all_ts_types = const.GRID_IO.TS_TYPES
    from pyaerocom.tstype import TsType
    lowest = TsType(ts_type)
    for freq in ts_types:
# =============================================================================
#         if not freq in all_ts_types:
#             raise ValueError('Invalid input, only valid ts_type codes are '
#                              'supported: {}'.format(all_ts_types))
# =============================================================================
        _temp = TsType(freq)
        if _temp < lowest:
            lowest = _temp
    return lowest.val

def get_highest_resolution(ts_type, *ts_types):
    """Get the highest resolution from several ts_type codes
    
    Parameters
    ----------
    ts_type : str
        first ts_type
    *ts_types
        one or more additional ts_type codes
    
    Returns
    -------
    str
        the ts_type that corresponds to the highest resolution
        
    Raises
    ------
    ValueError
        if one of the input ts_type codes is not supported
    """
    all_ts_types = const.GRID_IO.TS_TYPES
    highest = ts_type
    for freq in ts_types:
        if not freq in all_ts_types:
            raise ValueError('Invalid input, only valid ts_type codes are '
                             'supported: {}'.format(all_ts_types))
        elif all_ts_types.index(highest) > all_ts_types.index(freq):
            highest = freq
    return highest

def isnumeric(val):
    """Check if input value is numeric
    
    Parameters
    ----------
    val
        input value to be checked
    
    Returns
    -------
    bool 
        True, if input value corresponds to a range, else False.
    """
    from numbers import Number
    if isinstance(val, Number):
        return True
    return False
    
def isrange(val):
    """Check if input value corresponds to a range
    
    Checks if input is list, or array or tuple with 2 entries, or alternatively
    a slice that has defined start and stop and has set step to None.

    Note
    ----
    No check is performed, whether first entry is smaller than second entry if
    all requirements for a range are fulfilled.
    
    Parameters
    ----------
    val
        input value to be checked
    
    Returns
    -------
    bool 
        True, if input value corresponds to a range, else False.
    """
    if isinstance(val, (list, np.ndarray, tuple)):
        if len(val) == 2:
            return True
        return False
    elif isinstance(val, slice):
        if val.step is not None or val.start is None or val.stop is None:
            return False
        return True
    return False
       
def merge_station_data(stats, var_name, pref_attr=None, 
                       sort_by_largest=True, fill_missing_nan=True, 
                       **add_meta_keys):
    """Merge multiple StationData objects (from one station) into one instance
    
    Note
    ----
    - all input :class:`StationData` objects need to have same attributes\
       ``station_name``, ``latitude``, ``longitude`` and ``altitude``
    
    Parameters
    ----------
    stats : list
        list containing :class:`StationData` objects (note: all of these 
        objects must contain variable data for the specified input variable)
    var_name : str
        data variable name that is to be merged
    pref_attr 
        optional argument that may be used to specify a metadata attribute
        that is available in all input :class:`StationData` objects and that
        is used to order the input stations by relevance. The associated values
        of this attribute need to be sortable (e.g. revision_date). This is 
        only relevant in case overlaps occur. If unspecified the relevance of 
        the stations is sorted based on the length of the associated data 
        arrays.
    sort_by_largest : bool
        if True, the result from the sorting is inverted. E.g. if 
        ``pref_attr`` is unspecified, then the stations will be sorted based on
        the length of the data vectors, starting with the shortest, ending with
        the longest. This sorting result will then be inverted, if 
        ``sort_by_largest=True``, so that the longest time series get's highest
        importance. If, e.g. ``pref_attr='revision_date'``, then the stations 
        are sorted by the associated revision date value, starting with the 
        earliest, ending with the latest (which will also be inverted if 
        this argument is set to True)
    fill_missing_nan : bool
        if True, the resulting time series is filled with NaNs. NOTE: this 
        requires that information about the temporal resolution (ts_type) of
        the data is available in each of the StationData objects.
    """    
    if isinstance(var_name, list):
        if len(var_name) > 1:
            raise NotImplementedError('Merging of multivar data not yet possible')
        var_name = var_name[0]
        
    # make sure the data is provided as pandas.Series object
    is_3d, has_errs = False, False
    for stat in stats:
        if not var_name in stat:
            raise DataCoverageError('All input station must contain {} data'
                                    .format(var_name))
        elif pref_attr is not None and not pref_attr in stat:
            raise MetaDataError('Cannot sort station relevance by attribute {}. '
                                'At least one of the input stations does not '
                                'contain this attribute'.format(pref_attr))
        elif not isinstance(stat[var_name], pd.Series):
            try:
                stat._to_ts_helper(var_name)
            except Exception as e:
                raise ValueError('Data needs to be provided as pandas Series in '
                                 'individual station data objects. Attempted to'
                                 'convert but failed with the following '
                                 'exception: {}'.format(repr(e)))
        elif fill_missing_nan:
            try:
                stat.get_var_ts_type(var_name)
            except MetaDataError:
                raise MetaDataError('Cannot merge StationData objects: one or '
                                    'more of the provided objects does not '
                                    'provide information about the ts_type of '
                                    'the {} data, which is required when input '
                                    'arg. fill_missing_nan is True.'.format(var_name))
        if stat.check_if_3d(var_name):
            is_3d = True
        elif is_3d:
            raise ValueError('Merge error: some of the input stations contain '
                             'altitude info (suggesting profile data), others '
                             'not.')
        if var_name in stat.data_err:
            has_errs = True
            
    if not is_3d:
        if pref_attr is not None:
            stats.sort(key=lambda s: s[pref_attr])
        else:
            stats.sort(key=lambda s: len(s[var_name].dropna()))
        
        if sort_by_largest:
            stats = stats[::-1]

        # remove first station from the list
        merged = stats.pop(0)
            
        for i, stat in enumerate(stats): 
            merged.merge_other(stat, var_name, **add_meta_keys)
    else:
        from xarray import DataArray
        dtime = []
        for stat in stats:
            _t = stat[var_name].index.unique()
            if not len(_t) == 1:
                raise NotImplementedError('So far, merging of profile data '
                                          'requires that profile values are '
                                          'sampled at the same time')
            dtime.append(_t[0])
        tidx = pd.DatetimeIndex(dtime)
        
        # AeroCom default vertical grid
        vert_grid = const.make_default_vert_grid()
        _data = np.ones((len(vert_grid), len(tidx))) * np.nan
        if has_errs:
            _data_err = np.ones((len(vert_grid), len(tidx))) * np.nan
        
        for i, stat in enumerate(stats):
            #print(stat[var_name].values)
            if i == 0:
                merged = stat
            else:
                merged.merge_meta_same_station(stat, **add_meta_keys)

            _data[:, i] = np.interp(vert_grid, stat['altitude'], 
                                    stat[var_name].values)
            
            if has_errs:
                try:
                    _data_err[:, i] = np.interp(vert_grid, 
                                                stat['altitude'], 
                                                stat.data_err[var_name])
                except:
                    pass
        _coords = {'time'     : tidx,
                   'altitude' : vert_grid}
        
        d = DataArray(data=_data, coords=_coords, 
                      dims=['altitude', 'time'], name=var_name)
        d = d.sortby('time')
        merged[var_name] = d
        merged.dtime = d.time
        merged.altitude = d.altitude
    
    if fill_missing_nan:
        try:
            merged.insert_nans_timeseries(var_name)
        except Exception as e:
            const.print_log.warning('Could not insert NaNs into timeseries of '
                                    'variable {} after merging stations. '
                                    'Reason: {}'.format(var_name, repr(e)))
            
    merged['stat_merge_pref_attr'] = pref_attr
    return merged

def _get_pandas_freq_and_loffset(freq):
    """Helper to convert resampling info"""
    if freq in TS_TYPE_TO_PANDAS_FREQ:
        freq = TS_TYPE_TO_PANDAS_FREQ[freq]
    loffset = None
    if freq in PANDAS_RESAMPLE_OFFSETS:
        loffset = PANDAS_RESAMPLE_OFFSETS[freq]
    return (freq, loffset)

def resample_timeseries(s, freq, how='mean', min_num_obs=None):
    """Resample a timeseries (pandas.Series)
    
    Parameters
    ----------
    s : Series
        time series instance
    freq : str
        new temporal resolution (can be pandas freq. string, or pyaerocom
        ts_type)
    how : str
        choose from mean or median
    min_num_obs : :obj:`int`, optional
        minimum number of observations required per period (when downsampling).
        E.g. if input is in daily resolution and freq is monthly and 
        min_num_obs is 10, then all months that have less than 10 days of data
        are set to nan.
    
    Returns
    -------
    Series
        resampled time series object
    """
    freq, loffset = _get_pandas_freq_and_loffset(freq)    
    resampler = s.resample(freq, loffset=loffset)
    if min_num_obs is None:
        data = resampler.agg(how)
    else:
        df = resampler.agg([how, 'count'])
        const.logger.info(freq, min_num_obs)
        const.logger.info('before mean', df[how].mean())
        invalid = df['count'] < min_num_obs
        const.logger.info(len(invalid), invalid.sum())
        df[how][invalid] = np.nan
        const.logger.info('after mean', df[how].mean())
        data = df[how]

    return data.loc[s.index[0]:s.index[-1]]

def resample_time_dataarray(arr, freq, how='mean', min_num_obs=None):
    """Resample the time dimension of a :class:`xarray.DataArray`
    
    Note
    ----
    The dataarray must have a dimension coordinate named "time"
    
    Parameters
    ----------
    arr : DataArray
        data array to be resampled
    freq : str
        new temporal resolution (can be pandas freq. string, or pyaerocom
        ts_type)
    how : str
        choose from mean or median
    min_num_obs : :obj:`int`, optional
        minimum number of observations required per period (when downsampling).
        E.g. if input is in daily resolution and freq is monthly and 
        min_num_obs is 10, then all months that have less than 10 days of data
        are set to nan.
    
    Returns
    -------
    DataArray
        resampled data array object
    
    Raises
    ------
    IOError
        if data input `arr` is not an instance of :class:`DataArray`
    DataDimensionError
        if time dimension is not available in dataset
    """
    
    if not isinstance(arr, xray.DataArray):
        raise IOError('Invalid input for arr: need DataArray, got {}'.format(type(arr)))
    elif not 'time' in arr.dims:
        raise DataDimensionError('Cannot resample time: input DataArray has '
                                 'no time dimension')

    from pyaerocom.tstype import TsType
    from pyaerocom.time_config import XARR_TIME_GROUPERS
    to = TsType(freq)
    pd_freq=to.to_pandas()
    invalid = None
    if min_num_obs is not None:
        if not pd_freq in XARR_TIME_GROUPERS:
            raise ValueError('Cannot infer xarray grouper for ts_type {}'
                             .format(to.val))
        gr = XARR_TIME_GROUPERS[pd_freq]
        # 2D mask with shape of resampled data array
        invalid = arr.groupby('time.{}'.format(gr)).count(dim='time') < min_num_obs

    freq, loffset = _get_pandas_freq_and_loffset(freq)    
    arr = arr.resample(time=pd_freq, loffset=loffset).mean(dim='time')
    if invalid is not None:
        arr.data[invalid.data] = np.nan
    return arr
    
def same_meta_dict(meta1, meta2, ignore_keys=['PI'], 
                   num_keys=NUM_KEYS_META, num_rtol=1e-2):
    """Compare meta dictionaries
    
    Parameters
    ----------
    meta1 : dict
        meta dictionary that is to be compared with ``meta2``
    meta2 : dict
        meta dictionary that is to be compared with ``meta1``
    ignore_keys : list
        list containing meta keys that are supposed to be ignored
    num_keys : keys that contain numerical values
    num_rtol : float
        relative tolerance level for comparison of numerical values
        
    Returns
    -------
    bool 
        True, if dictionaries are the same, else False
    """
    if not meta1.keys() == meta2.keys():
        return False
    for k, v in meta1.items():
        if k in ignore_keys:
            continue
        elif k in num_keys:
            if not np.isclose(v, meta2[k], rtol=num_rtol):
                return False
        elif isinstance(v, dict):
            if not same_meta_dict(v, meta2[k]):
                return False
        else:
            if not v == meta2[k]:
                return False
    return True
            
def str_to_iris(key, **kwargs):
    """Mapping function that converts strings into iris analysis objects
    
    Please see dictionary ``STR_TO_IRIS`` in this module for valid definitions
    
    Parameters
    ----------
    key : str
        key of :attr:`STR_TO_IRIS` dictionary
        
    Returns
    -------
    obj
        corresponding iris analysis object (e.g. Aggregator, method)
    """
    key = key.lower()
    if not key in STR_TO_IRIS:
        raise KeyError("No iris.analysis object available for key %s, please "
                       "choose from %s" %(key, STR_TO_IRIS.keys()))
    val = STR_TO_IRIS[key]
    if callable(val):
        return val(**kwargs)
    return val

def to_pandas_timestamp(value):
    """Convert input to instance of :class:`pandas.Timestamp`"""
    if isinstance(value, pd.Timestamp):
        return value
    elif isinstance(value, (str, np.datetime64, datetime, date)):
        return pd.Timestamp(value)
    else:
        try:
            numval = int(value)
            if not 0 <= numval <= 10000:
                raise ValueError('Could not infer valid year from numerical '
                                 'time input')
            return pd.Timestamp(str(numval))
        except Exception as e:
            raise ValueError('Failed to convert {} to Timestamp: {}'
                             .format(value, repr(e)))    
    
def to_datetime64(value):
    """Convert input value to numpy.datetime64 
    
    Parameters
    ----------
    value
        input value that is supposed to be converted, needs to be either str, 
        datetime.datetime, pandas.Timestamp or an integer specifying the 
        desired year.
        
    Returns
    -------
    datetime64
        input timestamp converted to datetime64
    """
    if isinstance(value, np.datetime64):
        return value
    else:
        try:
            return to_pandas_timestamp(value).to_datetime64()
        except Exception as e:
            raise ValueError('Failed to convert {} to datetime64 object'
                             'Error: {}'.format(value, repr(e)))
  
def is_year(val):
    """Check if input is / may be year
    
    Parameters
    ----------
    val
        input that is supposed to be checked    
    
    Returns
    -------
    bool
        True if input is a number between -2000 and 10000, else False
    """
    try:
        if -2000 < int(val) < 10000:
            return True
        raise Exception
    except:
        return False
  
def _check_climatology_timestamp(t):
    if isnumeric(t) and t == 9999:
        return pd.Timestamp('1-1-2222')
    elif isinstance(t, np.datetime64):
        tstr = str(t)
        if tstr.startswith('9999'):
            return pd.Timestamp(tstr.replace('9999', '2222'))
    elif isinstance(t, str) and '9999' in t:
        return pd.Timestamp(t.replace('9999', '2222'))
    elif isinstance(t, datetime) and t.year == 9999:
        return pd.Timestamp(t.replace(year=2222))
    raise ValueError('Failed to identify {} as climatological timestamp...'
                     .format(t))
    
def start_stop(start, stop=None):
    """Create pandas timestamps from input start / stop values
    
    Note
    ----
    If input suggests climatological data in AeroCom format (i.e. year=9999) 
    then the year is converted to 2222 instead since pandas cannot handle 
    year 9999.
    
    Parameters
    -----------
    start
        start time (any format that can be converted to pandas.Timestamp)
    stop
        stop time (any format that can be converted to pandas.Timestamp)
    
    Returns
    -------
    pandas.Timestamp
        start timestamp
    pandas.Timestamp
        stop timestamp
    
    Raises
    ------
    ValueError
        if input cannot be converted to pandas timestamps
    """
    isclim = False
    try:
        start = to_pandas_timestamp(start)
    except pd.errors.OutOfBoundsDatetime: # probably climatology
        start = _check_climatology_timestamp(start)
        isclim = True
        
    if stop is None:
        if isclim:
            yr = 2222
        else:
            yr = start.year
        stop = to_pandas_timestamp('{}-12-31 23:59:59'.format(yr))
    else:
        try:
            stop = to_pandas_timestamp(stop)
        except pd.errors.OutOfBoundsDatetime:
            stop = _check_climatology_timestamp(stop)
    return (start, stop)

def datetime2str(time, ts_type=None):
    conv = TS_TYPE_DATETIME_CONV[ts_type]
    if is_year(time):
        return str(time)
    try:
        time = to_pandas_timestamp(time).strftime(conv)
    except pd.errors.OutOfBoundsDatetime:
        const.print_log.warning('Failed to convert time {} to string'.format(time))
        pass
    return time

def start_stop_str(start, stop=None, ts_type=None):
    
    conv = TS_TYPE_DATETIME_CONV[ts_type]
    if is_year(start) and stop is None:
        return str(start)
    start, stop = start_stop(start, stop)
    start_str = start.strftime(conv)
    stop_str = stop.strftime(conv)
    if stop_str != start_str:
        return '{}-{}'.format(start_str, stop_str)
    return start_str

def start_stop_from_year(year):
    """Create start / stop timestamp from year
    
    Parameters
    ----------
    year : int
        the year for which start / stop is to be instantiated
        
    Returns
    -------
    tuple
        2-element tuple containing
        
        - :obj:`pandas.Timestamp`: start timestamp
        - :obj:`pandas.Timestamp`: end timestamp
    """
    start = to_pandas_timestamp(year)
    stop = to_pandas_timestamp('{}-12-31 23:59:59'.format(year))
    return (start, stop)
            
def to_datestring_YYYYMMDD(value):
    """Convert input time to string with format YYYYMMDD
    
    Parameters
    ----------
    value
        input time, may be string, datetime, numpy.datetime64 or 
        pandas.Timestamp
        
    Returns
    -------
    str
        input formatted to string YYYYMMDD
        
    Raises
    ------
    ValueError
        if input is not supported
    """
    if isinstance(value, str) and len(value, 8):
        logger.info('Input is already string containing 8 chars. Assuming it '
                    'is in the right format and returning unchanged')
        return value
    try:
        return to_pandas_timestamp(value).strftime('%Y%m%d')
    except Exception as e:
        raise ValueError('Invalid input, need str, datetime, numpy.datetime64 '
                         'or pandas.Timestamp. Error: {}'.format(repr(e)))
    
def cftime_to_datetime64(times, cfunit=None, calendar=None):
    """Convert numerical timestamps with epoch to numpy datetime64
    
    This method was designed to enhance the performance of datetime conversions
    and is based on the corresponding information provided in the cftime 
    package (`see here <https://github.com/Unidata/cftime/blob/master/cftime/
    _cftime.pyx>`__). Particularly, this object does, what the :func:`num2date` 
    therein does, but faster, in case the time stamps are not defined on a non
    standard calendar.
    
    Parameters
    ----------
    times : :obj:`list` or :obj:`ndarray` or :obj:`iris.coords.DimCoord`
        array containing numerical time stamps (relative to basedate of 
        ``cfunit``). Can also be a single number.
    cfunit : :obj:`str` or :obj:`Unit`, optional
        CF unit string (e.g. day since 2018-01-01 00:00:00.00000000 UTC) or
        unit. Required if `times` is not an instance of 
        :class:`iris.coords.DimCoord`
    calendar : :obj:`str`, optional
        string specifying calendar (only required if ``cfunit`` is of type
        ``str``).
        
    Returns
    -------
    ndarray
        numpy array containing timestamps as datetime64 objects
        
    Raises
    ------
    ValueError
        if cfunit is ``str`` and calendar is not provided or invalid, or if 
        the cfunit string is invalid
        
    Example
    -------
    
    >>> cfunit_str = 'day since 2018-01-01 00:00:00.00000000 UTC'
    >>> cftime_to_datetime64(10, cfunit_str, "gregorian")
    array(['2018-01-11T00:00:00.000000'], dtype='datetime64[us]')
    """
    if isinstance(times, iris.coords.DimCoord): #special case
        times, cfunit = times.points, times.units
    try:
        len(times)
    except:
        times = [times]
    if isinstance(cfunit, str):
        if calendar is None:
            raise ValueError("Require specification of calendar for "
                             "conversion into datetime64 objects")
        cfunit = Unit(cfunit, calendar) #raises Error if calendar is invalid
    if not isinstance(cfunit, Unit):
        raise ValueError("Please provide cfunit either as instance of class "
                         "cf_units.Unit or as a string")
    calendar = cfunit.calendar
    basedate = cfunit.num2date(0)
    if ((calendar == 'proleptic_gregorian' and basedate.year >= MINYEAR) or 
        (calendar in ['gregorian','standard'] and basedate > GREGORIAN_BASE)):
        # NOTE: changed on 9 July 2018 by jgliss due to error (kernel died)
        # after update of dependencies (cf_units). Attribute name does not
        # work anymore...
        cfu_str = cfunit.origin #cfunit.name
        
        res = cfu_str.split()[0].lower()
        if res in microsec_units:
            tstr = "us"
        elif res in millisec_units:
            tstr = "ms"
        elif res in sec_units:
            tstr = "s"
        elif res in min_units:
            tstr = "m"
        elif res in hr_units:
            tstr = "h"
        elif res in day_units:
            tstr = "D"
        else:
            raise ValueError('unsupported time units')
        
        basedate = np.datetime64(basedate)
        dt = np.asarray(np.asarray(times), dtype="timedelta64[%s]" %tstr)
        return basedate + dt
    else:
        return np.asarray([np.datetime64(t) for t in cfunit.num2date(times)])

def get_constraint(lon_range=None, lat_range=None, 
                   time_range=None, meridian_centre=True):
    """Function that creates an :class:`iris.Constraint` based on input
    
    Note
    ----
    Please be aware of the definition of the longitudes in your data when 
    cropping within the longitude dimension. The longitudes in your data may be 
    defined either from **-180 <= lon <= 180** (pyaerocom standard) or from 
    **0 <= lon <= 360**. In the former case (-180 -> 180) you can leave the 
    additional input parameter ``meridian_centre=True`` (default). 
    
    Parameters
    ----------
    lon_range : :obj:`tuple`, optional
        2-element tuple containing longitude range for cropping
        Example input to crop around meridian: `lon_range=(-30, 30)`
    lat_range : :obj:`tuple`, optional
        2-element tuple containing latitude range for cropping. 
    time_range : :obj:`tuple`, optional
        2-element tuple containing time range for cropping. Allowed data
        types for specifying the times are 
        
            1. a combination of 2 :class:`pandas.Timestamp` instances or 
            2. a combination of two strings that can be directly converted\
            into :class:`pandas.Timestamp` instances (e.g.\
            `time_range=("2010-1-1", "2012-1-1")`) or
            3. directly a combination of indices (:obj:`int`). 
    meridian_centre : bool
        specifies the coordinate definition range of longitude array. If True, 
        then -180 -> 180 is assumed, else 0 -> 360
        
    Returns
    -------
    iris.Constraint
        the combined constraint from all valid input parameters
    
    Examples
    --------
    The following example shows how to crop over the meridian
    
    >>> from pyaerocom.helpers import get_constraint
    >>> from pyaerocom.io.fileconventions import FileConventionRead
    >>> from iris import load
    >>> from pyaerocom.io.testfiles import get
    >>> files = get()
    >>> fname = files['models']['aatsr_su_v4.3']
    >>> convention = FileConventionRead().from_file(fname)
    >>> meta_info = convention.get_info_from_file(fname)
    >>> for k, v in meta_info.items(): print(k, v)
    year 2008
    var_name od550aer
    ts_type daily
    >>> cubes = load(fname)
    >>> lons = cubes[0].coord("longitude").points
    >>> meridian_centre = True if lons.max() > 180 else False
    >>> year = meta_info["year"]
    >>> c = get_constraint(lon_range=(50, 150), 
    ...                    lat_range=(20, 60), 
    ...                    time_range=("%s-02-05" %year, "%s-02-25" %year))
    >>> cube_crop = cubes.extract(c)[0]
    >>> cube_crop.shape
    (21, 40, 100)
    """
    constraints = []
    if lon_range is not None:
        constraints.append(get_lon_rng_constraint(lon_range, meridian_centre))
    if lat_range is not None:
        constraints.append(get_lat_rng_constraint(lat_range))
    if time_range is not None:
        constraints.append(get_time_rng_constraint(*time_range))
    if len(constraints) > 0:
        c = constraints[0]
        for cadd in constraints[1:]:
            c = c & cadd
    return c
    
def get_lat_rng_constraint(lat_range):
    """Create latitude constraint based on input range
    
    Parameters
    ----------
    lat_range : tuple
        2-element tuple specifying latitude range
    
    Returns
    -------
    iris.Constraint
        the corresponding iris.Constraint instance
        
    """
    return iris.Constraint(latitude=lambda v: lat_range[0] <= v <= lat_range[1])
    
def get_lon_rng_constraint(lon_range, meridian_centre=True):
    """Create longitude constraint based on input range

    Parameters
    ----------
    lon_range : tuple
        2-element tuple containing from left -> right end of range
    meridian_centre : bool
        specifies the coordinate definition range of longitude array of the 
        data to be cropped. If True, then -180 -> 180 is assumed, else 0 -> 360
    
    Returns
    -------
    iris.Constraint
        the corresponding iris.Constraint instance 
    
    Raises
    ------
    ValueError
        if first coordinate in lon_range equals or exceeds second
    LongitudeConstraintError
        if the input implies cropping over border of longitude array
        (e.g. 160 -> - 160 if -180 <= lon <= 180).

    Example
    -------
    >>> from pyaerocom.io.testfiles import get
    >>> from pyaerocom import GriddedData
    >>> files = get()
    >>> data = GriddedData(files['models']['aatsr_su_v4.3'], var_name="od550aer")
    >>> c = get_lon_rng_constraint(lon_range=(170, -160), meridian_centre=True)
    Traceback (most recent call last):
     ...
    ValueError: Left coordinate must exceed right coordinate
    >>> c = get_lon_rng_constraint(lon_range=(-30, 30), meridian_centre=True)
    >>> data_crop = data.extract(c)
    >>> assert data_crop.grid.shape == (366, 180, 60)
    """
    left, right = lon_range
    if left == right:
        raise ValueError("the specified values are equal")
    elif left > right:
        raise ValueError("Left coordinate must exceed right coordinate")
    if meridian_centre:
        left, right = (left+180)%360-180, (right+180)%360-180
    else:
        left, right = left%360, right%360
    if left > right:
        msg = ("Cannot crop over right border of longitude range")
        raise LongitudeConstraintError(msg)
    return iris.Constraint(longitude=lambda v: left <= v <= right)

def get_time_rng_constraint(start, stop):
    """Create iris.Constraint for data extraction along time axis
    
    Parameters
    ----------
    start : :obj:`Timestamp` or :obj:` str`
        start time of desired subset. If string, it must be convertible 
        into :class:`pandas.Timestamp` (e.g. "2012-1-1")
    stop : :obj:`Timestamp` or :obj:` str`
        start time of desired subset. If string, it must be convertible 
        into :class:`pandas.Timestamp` (e.g. "2012-1-1")
    
    Returns
    -------
    iris.Constraint
        iris Constraint instance that can, e.g., be used as input for
        :func:`pyaerocom.griddeddata.GriddedData.extract`
    """
    if not isinstance(start, pd.Timestamp):
        start = pd.Timestamp(start)
    if not isinstance(stop, pd.Timestamp):
        stop = pd.Timestamp(stop)
        
    t_lower = iris.time.PartialDateTime(year=start.year,
                                        month=start.month,
                                        day=start.day)
    t_upper = iris.time.PartialDateTime(year=stop.year,
                                        month=stop.month,
                                        day=stop.day)
    
    return iris.Constraint(time=lambda cell: t_lower <= cell <= t_upper)

if __name__=="__main__":
    print(get_lowest_resolution('yearly', 'daily', 'monthly'))
    print(get_highest_resolution('yearly', 'daily', 'monthly'))
    
    print(infer_time_resolution([np.datetime64('2010-01-01'),
                                 np.datetime64('2010-01-02'),
                                 np.datetime64('2010-01-05'),
                                 np.datetime64('2010-10-15')]))


# -*- coding: utf-8 -*-

"""
High level test methods that check AasEtAl time series data for selected stations
Created on June 12 2019.

@author: hannas@met.no
"""
import pytest
import numpy.testing as npt
import numpy as np
import pandas as pd

import datetime
import matplotlib.pyplot as plt
from scipy.stats import kendalltau
from scipy.stats.mstats import theilslopes

import pyaerocom as pya
from pyaerocom.test.settings import lustre_unavail, TEST_RTOL, test_not_working
from pyaerocom.io.read_aasetal import ReadSulphurAasEtAl
from pyaerocom.units_helpers import convert_unit, convert_unit_back

# (from to unit)
UNITCONVERSION = {'concso2':   ('ug S/m3', 'ug m-3'), 
                  'concso4':   ('ug S/m3', 'ug m-3'), 
                  'wetso4':    ('kg S/ha', 'kg m-2'),  #  s-1
                  'concso4pr': ('mg S/L',   'g m-3')
                  }
    
TRUE_SLOPE = {'N-America': {"concso4": {'2000–2010': -3.03,
                                          '2000–2015': -3.15},

                             "wetso4": {'2000–2010': -2.30,
                                        '2000–2015': -2.78},

                             'concso2': {'2000–2010': -4.55,
                                          '2000–2015':-4.69
                                          }},

                   'Europe': {'concso4': {'2000–2010':  -2.86,
                                           '2000–2015':  -2.67},

                              'wetso4': {'2000–2010': -3.85,
                                         '2000–2015': -3.40,
                                         },
                              'concso2': {'2000–2010': -4.23,
                                           '2000–2015':-3.89
                                           },
                              },
                   }

TRUE_NR_STATIONS = {'N-America': {"concso4": {'2000–2010': 227,
                                               '2000–2015': 218},
                              
                                     "wetso4": {'2000–2010': 226,
                                                '2000–2015': 215
                                                },
        
                                     'concso2': {'2000–2010': 78,
                                                  '2000–2015': 77
                                                  }},

                       'Europe': {'concso4': {'2000–2010': 43,
                                               '2000–2015': 36},

                                  'wetso4': {'2000–2010': 73,
                                             '2000–2015': 67
                                             },
                                  'concso2': {'2000–2010': 51,
                                               '2000–2015': 47
                                               },
                                  },
                       }

def _get_season_current(mm,yyyy):
    if mm in [3,4,5]:
        s = 'spring-'+str(int(yyyy))
    if mm in [6,7,8]:
        s = 'summer-'+str(int(yyyy))
    if mm in [9,10,11]:
        s = 'autumn-'+str(int(yyyy))
    if mm in [12]:
        s = 'winter-'+str(int(yyyy))
    if mm in [1,2]:
        s = 'winter-'+str(int(yyyy-1))
    return s

def _mid_season_current(seas, yr):
    """ Calculated the mid date for a season.
    
    From original trends interface developed by A. Mortier.
    
    Parameters
    ------------------
    seas : str
        Season
        
    yr : int
        year
    """
    if seas=='spring':
        date = datetime.datetime(yr,4,15)
    if seas=='summer':
        date = datetime.datetime(yr,7,15)
    if seas=='autumn':
        date = datetime.datetime(yr,10,15)
    if seas=='winter':
        date = datetime.datetime(yr-1,1,15)
    if seas=='all':
        date = datetime.datetime(yr,6,15)
    return date

def compute_trends_current(s_monthly, periods, only_yearly=True):
    """ Compute trends for station. 
    
    Used in the trends interface to compute the trends before. 05.06.19.

    Slightly modified code from original trends interface developed by
    A. Mortier.

    Parameters
    ------------------------

    s_monthly : pd.DataFrame
        Dataframe containing montly values of data.
    
    periods : list[str]
        List containing periods.

    only_yearly : boolean
    
    Returns 
    ------------------------
    data : dict
        Dictionary containing the trends.

    Main changes applied:
        - Keep NaNs

    """
    
    d = dict(month=s_monthly.index.month,
             year=s_monthly.index.year,
             value=s_monthly.values)

    mobs = pd.DataFrame(d)
    mobs['season'] = mobs.apply(lambda row: _get_season_current(row['month'],
                                                                row['year']), 
                                                                axis=1)
    # drop rows where value = nan.
    mobs = mobs.dropna(subset=['value'])

    # trends with yearly and seasonal averages
    seasons = ['spring', 'summer', 'autumn', 'winter', 'all']
    yrs = np.unique(mobs['year'])
    #print('yrs {}'.format(yrs))
    data = {}
    
    # added to minimize the computation
    for i, seas in enumerate(seasons):
        if only_yearly and not seas == 'all':
            continue
        # initialize seasonal object
        data[seas] = {'date': [], 'jsdate': [], 'val': []}
        # filter the months
        for yr in yrs:
            if seas != 'all':
                catch = mobs[mobs['season'].str.contains(seas + '-' + str(yr))]
            else:
                catch = mobs[mobs['season'].str.contains('-' + str(yr))]
            date = _mid_season_current(seas, yr)

            data[seas]['date'].append(date)
            epoch = datetime.datetime(1970, 1, 1)
            data[seas]['jsdate'] = [(dat - epoch).total_seconds() * 1000 for dat in data[seas]['date']]

            # needs 4 seasons to compute seasonal average to avoid biases
            if (seas == 'all') & (len(np.unique(catch['season'].values)) < 4):
                data[seas]['val'].append(np.nan)
            else:
                #print(catch['value'])
                data[seas]['val'].append(np.nanmean(catch['value']))
        
        # trends for this season
        data[seas]['trends'] = {}   
        # filter period
        for period in periods:
            p0 = int(period[:4])
            p1 = int(period[5:])
            data[seas]['trends'][period] = {}
      
            # Mann-Kendall test
            x = np.array(data[seas]['jsdate'])
            y = np.array(data[seas]['val'])
            len_period = len(y)

            # works only on not nan values
            x = x[~np.isnan(y)] # Better ith np.isfinite()
            y = y[~np.isnan(y)]

            # filtering to the period limit
            jsp0 = (datetime.datetime(p0, 1, 1) - epoch).total_seconds() * 1000
            jsp1 = (datetime.datetime(p1, 12, 31) - epoch).total_seconds() * 1000
            
            y = y[(x >= jsp0) & (x <= jsp1)] 
            x = x[(x >= jsp0) & (x <= jsp1)]
            
            # Making sure there is at least 75% coverage in the data period.
            # and that we have more than two points
            if len(y)/len_period >= 0.75 and len(y) > 1: 
                # Kendall
                
                # TODO THIS IS WHERE YOU SHOULD ASK AUGUSTIN HOW THINGS 
                # SHOULD BE RESTRICTED BY KENTAL TAu
                
                [tau, pval] = kendalltau(x, y)
                #print('pval {}'.format(pval))
                data[seas]['trends'][period]['pval'] = pval
                
                if pval < 0.1:
                    # Theil slope
                    res = theilslopes(y, x, 0.9)
                    medslope, medintercept, lo_slope, up_slope = res
                    reg = medslope* np.asarray(x) + medintercept * np.ones(len(x))
                    slp = res[0] * 1000 * 60 * 60 * 24 * 365.25 / reg[0]  # slp per milliseconds to slp per year
                    data[seas]['trends'][period]['slp'] = slp * 100  # in percent
                    data[seas]['trends'][period]['reg0'] = reg[0]
                    data[seas]['trends'][period]['t0'] = x[0]
                    data[seas]['trends'][period]['n'] = len(y)
                else:
                    data[seas]['trends'][period]['pval'] = None
                    data[seas]['trends'][period]['slp'] = None
                    data[seas]['trends'][period]['reg0'] = None
                    data[seas]['trends'][period]['t0'] = None
                    data[seas]['trends'][period]['n'] = len(y) 
            else:
                data[seas]['trends'][period]['pval'] = None
                data[seas]['trends'][period]['slp'] = None
                data[seas]['trends'][period]['reg0'] = None
                data[seas]['trends'][period]['t0'] = None
                data[seas]['trends'][period]['n'] = len(y) 
    return data

    """
    def test_unitconversion_surface_conc():
        a = 10
        temp = unitconv_sfc_conc(a, 2)
        A = unitconv_sfc_conc_bck(temp, 2)
        assert np.abs(a - A) < 0.000001
    """

    """
    @test_not_working
    def test_unitconversion_wetdep():
        a = 10
        time = pd.Series(np.datetime64('2002-06-28'))
        temp = unitconv_wet_depo(a, time)
        A = unitconv_wet_depo_bck(temp, time)
        assert np.abs(a - A) < 0.000001
    """
def create_region(region):
    """ Small modifications to the one in the trends interface.

    Function to create regions.

    Parameters
    -----------------------
    region : str
        Name of existing region.

    Returns
    -----------------------------
        minLon, maxLon, minLat, maxLat : ndarray[int]
    """

    regions = {
        # 'US': {'minLat': 22.5, 'maxLat': 71.0, 'minLon': -167, 'maxLon': -59.6},
        'N-America': {'minLat': 15.0, 'maxLat': 72.0, 'minLon': -170.0, 'maxLon': -50.0},
        'S-America': {'minLat': -55.0, 'maxLat': 14.0, 'minLon': -85.0, 'maxLon': -33.0},
        'Europe': {'minLat': 34.2, 'maxLat': 67.2, 'minLon': -28.7, 'maxLon': 69.1},
        'East-Asia': {'minLat': -12.62, 'maxLat': 59.9, 'minLon': 97.0, 'maxLon': 161.19},
        'Africa': {'minLat': -41.03, 'maxLat': 33.08, 'minLon': -18.45, 'maxLon': 58.54},
        'India': {'minLat': 3.06, 'maxLat': 36.0, 'minLon': 65.92, 'maxLon': 96.0},
        'East-US': {'minLat': 30, 'maxLat': 45, 'minLon': -95, 'maxLon': -75},
        'Central-Europe': {'minLat': 40, 'maxLat': 55, 'minLon': -5, 'maxLon': 40},
        'Most-East-Asia': {'minLat': 25, 'maxLat': 45, 'minLon': 100, 'maxLon': 130},
        'World': {'minLat': -90, 'maxLat': 90, 'minLon': -180, 'maxLon': 180},
        '': {'minLat': 0, 'maxLat': 0, 'minLon': 0, 'maxLon': 0, 'zf': 0},
    }
    return regions[region]['minLon'], regions[region]['maxLon'], regions[region]['minLat'], regions[region]['maxLat']


def crop_data_to_region(ungridded, region = "Europe"):
    """ Function which filter out the stations in the correct region.

    Parameters
    -------------
    ungridded : :class:`UngriddedData`
        Ungridded data object which contain data from one variable (filter by meta is not implemented for several variables yet.)

    region : str
        Name of region, default is N-America.

    Returns
    -------
    croppped_ungridded : :class:`UngriddedData`
        Ungridded data object which only contain stations from the correct data.

    """
    valid_names = ['N-America', 'S-America', 'Europe', 'East-Asia', 'Africa', 
                   'India', 'East-US', 'Central-Europe', 'Most-East-Asia']
    
    if region in valid_names:
        minLon, maxLon, minLat, maxLat = create_region(region = region)
        croppped_ungridded = ungridded.filter_by_meta(latitude=(minLat, maxLat), 
                                                      longitude=(minLon, maxLon))
        return croppped_ungridded
    else:
        raise ValueError("{} is not a valid region. "
                         "Try 'N-America', 'S-America', 'Europe', 'East-Asia', 'Africa', 'India', 'East-US', "
                         "'Central-Europe', 'Most-East-Asia' ".format(region))

def years_to_string(start, stop):
    """Function to make it simpler to retrieve testdata from dictionary."""
    return "{}–{}".format(start, stop)

def years_from_periodstr(period):
    return [int(x) for x in period.split('–')]

"""
def convert_data(data, var):
    if var == "concso4":
        return unitconv_sfc_conc_bck(data, x=4)
    elif var == "concso2":
        unitconv_sfc_conc_bck(data, x=2)
    elif var == "wetso4":
        unitconv_wet_depo_bck(data)
    else:
        raise ValueError("{} is not a valid variable.".format(var))
"""

def calc_slope(ungridded, region, period, var):
    """
    Check if the number of stations present is only counting the ones with more 
    than 7 years of averages.
    
    :return: Boolean
    """
    # crop data to region
    start, stop = years_from_periodstr(period[0])
    regional = crop_data_to_region(ungridded, region = region)

    # crop to timeseries
    stations =  regional.to_station_data_all(var, start, stop, 'monthly')
    name = stations['station_name']
    #print('nbr names {}'.format(len(name)))
    slp_stations = []
    station_names = []

    for station in stations['stats']:
        #date = station['dtime']
        ts = station[var]
        s_monthly = convert_data(ts.values, var)
        # not pandas series now may need to convert it back to that
        s_monthly = ts
        # input to compute_trends_current period is supposed to be a list.
        data = compute_trends_current(s_monthly, period, only_yearly=True)
        slp = data['all']['trends'][period[0]]['slp']

        if slp is not None:
            slp_stations.append(slp)
            station_names.append(station.station_name)
            
            
    print('\n Percentage {}/{} = {}'.format(len(station_names), len(name), 
                                          len(station_names)/ len(name) ))
    
    max_name = station_names[np.argmax(slp_stations)] 
    slp_stations.remove(np.max(slp_stations))
    slp_stations.remove(np.max(slp_stations))
    nr_stations = len(slp_stations)
    return nr_stations, np.mean(slp_stations)

@lustre_unavail
def test_ungriddeddata():
    reader = ReadSulphurAasEtAl('GAWTADsubsetAasEtAl')
    data = reader.read()  # read all variables
    assert len(data.station_name) == 890
    assert data.shape == (416243, 12) 

@lustre_unavail
def test_reading_routines():
    """
    Read one station Yellowstone NP. Retrive station from ungridded data object, 
    convert unit back and compare this to the raw data from the file.    
    """
    
    files = ['/lustre/storeA/project/aerocom/aerocom1//AEROCOM_OBSDATA/PYAEROCOM/GAWTADSulphurSubset/data/monthly_so2.csv', 
             '/lustre/storeA/project/aerocom/aerocom1//AEROCOM_OBSDATA/PYAEROCOM/GAWTADSulphurSubset/data/monthly_so4_aero.csv', 
             '/lustre/storeA/project/aerocom/aerocom1//AEROCOM_OBSDATA/PYAEROCOM/GAWTADSulphurSubset/data/monthly_so4_precip.csv']
    
    df = pd.read_csv(files[0], sep=",", low_memory=False)
    subset = df[df.station_name == 'Yellowstone NP']
    vals = subset['concentration_ugS/m3'].astype(float).values
    ungridded = ReadSulphurAasEtAl().read(vars_to_retrieve = 'concso2')
    station = ungridded.to_station_data('Yellowstone NP', 'concso2')
    from_unit, to_unit = UNITCONVERSION['concso2']
    conv = convert_unit(data=station.concso2.values, from_unit = from_unit, 
                        to_unit = to_unit, var_name = 'concso2')
    
    #conv = unitconv_sfc_conc_bck(station.concso2.values, 2)
    assert (np.abs(conv - vals).sum() < 0.000001, 
            'Unconsistancy between reading a file and reading a station. '+
            'File: monthly_so2.csv. Station: Yellowstone NP. '
            +'Variable: concso2. ' ) 

    df = pd.read_csv(files[1], sep=",", low_memory=False)
    subset = df[df.station_name == 'Payerne']
    vals = subset['concentration_ugS/m3'].astype(float).values
    ungridded = ReadSulphurAasEtAl().read(vars_to_retrieve = 'concso4')
    station = ungridded.to_station_data('Payerne', 'concso4')
    
    from_unit, to_unit = UNITCONVERSION['concso4']
    conv = convert_unit(data=station.concso4.values, from_unit = from_unit, 
                        to_unit = to_unit, var_name = 'concso4')
    
    #conv = unitconv_sfc_conc_bck(station.concso4.values, 4)
    summen = np.abs(conv - vals).sum()
    print(summen)
    assert (summen < 0.000001, 
            'Unconsistancy between reading a file and reading a station. ' 
            +'File: monthly_so4_aero.csv. Station: Payerne. '
            +'Variable: concso4.')
    
    station_name = 'Abington (CT15)'
    df = pd.read_csv(files[2], sep=",", low_memory=False)
    subset = df[df.station_name == station_name]
    
    tconv = lambda yr, m : np.datetime64('{:04d}-{:02d}-{:02d}'.format(yr, m, 1), 's')
    dates_alt = [tconv(yr, m) for yr, m in zip(subset.year.values, subset.month.values)]
    subset['dtime'] = np.array(dates_alt)
    vals = subset['deposition_kgS/ha'].astype(float).values  
    print('Numbers of nans in original files {}.'.format(np.isnan(vals).sum()))
    
    ungridded = ReadSulphurAasEtAl().read(vars_to_retrieve = 'wetso4')
    station = ungridded.to_station_data(station_name, 'wetso4')
    #conv = unitconv_wet_depo_bck(station.wetso4.values, subset['dtime'], 'monthly').values
    conv = station.wetso4.values
    print('Numbers of nans in converted files {}.'.format(np.isnan(station.wetso4.values).sum()))
    
    
    summen = np.abs(conv - vals).sum()
    assert (summen < 0.00001, 'Unconsistancy between reading a file and reading a'
            +'station. File: monthly_so4_aero.csv. Station: {}. '
            +'Variable: wetso4.'.format(station_name) )
    print('Passed test on reading routines. ')
    return


def test_nbr_of_nans():
    
    """
    @lustre_unavail
    @test_not_working
    def test_article():
        regions = ['Europe', 'N-America']
        VARS = ['concso4', 'concso2', 'wetso4']
        periods = ['2000–2010', '2000–2015'] 
        # OBS (bad coding) problems with the heigfen. 
    
        for v in VARS:
            # can reuse the ungridded data object for all regions and periods.
            ungridded = ReadSulphurAasEtAl().read(vars_to_retrieve=v)
            for r in regions:
                for p in periods:
                    # true values
                    true_mean_slope = TRUE_SLOPE[r][v][p]
                    ns = TRUE_NR_STATIONS[r][v][p]
                    # computed values              
                    nr_stations, predicted_mean_slope = calc_slope(ungridded, r, [p], v)
                    print("REGION {}, variable {}, period {} ".format(r, v, p))
                    print("calc {} ,  true [nr stations] {}".format(nr_stations, ns))
                    print("calc_slope {}, true_slope   {}".format(
                                            np.around(predicted_mean_slope, 2), true_mean_slope))
                    
                    # Test similarity:
                    # TODO 1) check slope (calc_slope - true_slope) / true_slope x 100
                    # TODO 2) check that the number of stations are the same.
                    
                    #assert nr_stations == ns1
                    #assert nr_stations2 == ns2
                    #assert predicted_mean_slope1 == true_mean_slope1
                    #assert predicted_mean_slope2 == true_mean_slope2
    """
    pass

if __name__ == "__main__":
    #pya.change_verbosity('info')
    # import sys
    
    # OBS the three below works
    #test_unitconversion_surface_conc()
    #test_unitconversion_wetdep()
    #test_ungriddeddata()
    
    # TODO test article on hold 
    #test_reading_routines()
    #test_unitconversion_wetdep()
    reader = ReadSulphurAasEtAl('GAWTADsubsetAasEtAl')
    data = reader.read()  # read all variables
    print(data.shape)
    print(len(data.station_name))
# =============================================================================



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
import numpy as np
import numpy.testing as npt
from pyaerocom import UngriddedData
from pyaerocom.test.settings import lustre_unavail
from pyaerocom.io.test.test_read_aeronet_sunv3 import aeronetsunv3lev2_subset
from pyaerocom.exceptions import DataCoverageError

def test_init_shape():
    npt.assert_array_equal(UngriddedData().shape, (10000, 12))
    
    d1 = UngriddedData(num_points=2, add_cols=['bla', 'blub'])
    npt.assert_array_equal(d1.shape, (2, 14))
    
    d1.add_chunk(1112)
    
    npt.assert_array_equal(d1.shape, (1114, 14))
    

def test_coordinate_access():
    import string
    d = UngriddedData()
    
    stat_names = list(string.ascii_lowercase)
    lons = np.arange(len(stat_names))
    lats = np.arange(len(stat_names)) - 90
    alts = np.arange(len(stat_names)) * 13
    
    for i, n in enumerate(stat_names):
        d.metadata[i] = dict(data_id = 'testcase',
                             station_name = n, 
                             latitude = lats[i],
                             longitude = lons[i],
                             altitude = alts[i])
        
    import numpy.testing as npt
    
    npt.assert_array_equal(d.station_name, stat_names)
    npt.assert_array_equal(d.latitude, lats)
    npt.assert_array_equal(d.longitude, lons)
    npt.assert_array_equal(d.altitude, alts)
    
    case_ok = False
    try:
        d.to_station_data('a')
    except DataCoverageError:
        case_ok = True
        
    assert case_ok
    
    c = d.station_coordinates 
    npt.assert_array_equal(c['station_name'], stat_names)
    npt.assert_array_equal(c['latitude'], lats)
    npt.assert_array_equal(c['longitude'], lons)
    npt.assert_array_equal(c['altitude'], alts)
    
@lustre_unavail
def test_check_index_aeronet_subset(aeronetsunv3lev2_subset):
    aeronetsunv3lev2_subset._check_index()
    
if __name__=="__main__":
    test_init_shape()
    test_coordinate_access()